# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest import mock

import ddt

from tacker import context
from tacker import objects
from tacker.tests.unit.db.base import SqlTestCase
from tacker.tests.unit.objects import fakes
from tacker.tests import uuidsentinel


@ddt.ddt
class TestVnfLcmSubScriptions(SqlTestCase):

    def setUp(self):
        super(TestVnfLcmSubScriptions, self).setUp()
        self.context = context.get_admin_context()
        self.vnf_package = self._create_vnf_package()
        self.vnf_package_vnfd = self._create_and_upload_vnf_package_vnfd()
        self.subscription = self._create_subscription()

    def _create_vnf_package(self):
        vnfpkgm = objects.VnfPackage(context=self.context,
                                     **fakes.vnf_package_data)
        vnfpkgm.create()
        return vnfpkgm

    def _create_and_upload_vnf_package_vnfd(self):
        vnf_package = objects.VnfPackage(context=self.context,
                                         **fakes.vnf_package_data)
        vnf_package.create()

        vnf_pack_vnfd = fakes.get_vnf_package_vnfd_data(
            vnf_package.id, uuidsentinel.vnfd_id)

        vnf_pack_vnfd_obj = objects.VnfPackageVnfd(
            context=self.context, **vnf_pack_vnfd)
        vnf_pack_vnfd_obj.create()

        return vnf_pack_vnfd_obj

    @mock.patch.object(objects.vnf_lcm_subscriptions,
                       '_vnf_lcm_subscriptions_create')
    def _create_subscription(self, mock_vnf_lcm_subscriptions_create):
        filter = fakes.filter
        mock_vnf_lcm_subscriptions_create.return_value = \
            '{\
                "filter": "{"operationStates": ["COMPLETED"],\
                "vnfInstanceNames": ["xxxxxxxxxxxxxxxxxx"],\
                "operationTypes": ["INSTANTIATE"],\
                "vnfdIds": ["405d73c7-e964-4c8b-a914-41478ccd7c42"],\
                "vnfProductsFromProviders": [{\
                    "vnfProvider": "x2x", \
                    "vnfProducts": [{\
                        "vnfProductName": "x2xx", \
                        "versions": [{\
                            "vnfSoftwareVersion": "xx2XX", \
                            "vnfdVersions": ["ss2"]\
                        }]\
                    }]\
                }, \
                {\
                    "vnfProvider": "z2z",\
                    "vnfProducts": [{\
                        "vnfProductName": "z2zx", \
                        "versions": [{\
                            "vnfSoftwareVersion": "xx3XX",\
                            "vnfdVersions": \
                            ["s3sx", "s3sa"]\
                        }\
                    ]},\
                    {\
                        "vnfProductName": "zz3ex",\
                        "versions": [{\
                            "vnfSoftwareVersion": "xxe3eXz",\
                            "vnfdVersions": ["ss3xz", "s3esaz"]\
                        },\
                        {\
                            "vnfSoftwareVersion": "xxeeeXw", \
                            "vnfdVersions": ["ss3xw", "ss3w"]\
                        }]\
                    }]\
                }],\
                "notificationTypes": [\
                    "VnfLcmOperationOccurrenceNotification"],\
                "vnfInstanceIds": ["fb0b9a12-4b55-47ac-9ca8-5fdd52c4c07f"]}",\
                "callbackUri": "http://localhost/xxx",\
                "_links": {\
                    "self": {\
                        "href":\
                            "http://localhost:9890//vnflcm/v1/subscriptions\
                            /530a3c43-043a-4b84-9d65-aa0df49f7ced"\
                    }\
                },\
                "id": "530a3c43-043a-4b84-9d65-aa0df49f7ced",\
                "tenant_id": "8cd59fef-397e-4bec-8621-0e69225434f0"\
            }'

        subscription_obj = \
            objects.vnf_lcm_subscriptions.LccnSubscriptionRequest(
                context=self.context, **fakes.subscription_data)
        subscription_obj.create(filter)

        return subscription_obj

    @mock.patch.object(objects.vnf_lcm_subscriptions,
                       '_vnf_lcm_subscriptions_id_get')
    @ddt.data(
        fakes.filter,
        {"notificationTypes": ["VnfIdentifierCreationNotification"]}
    )
    def test_create(self, filter, mock_vnf_lcm_subscriptions_id_get):
        """Test that the `create` method returns a created filter

        Also, test `_vnf_lcm_subscriptions_create` method, which is called
        from the `create` method, works correctly and a created filter will be
        returned.

        Case 1: All available parameters are set
        Case 2: Only the required parameter is set
        """

        mock_vnf_lcm_subscriptions_id_get.return_value = None
        subscription_obj = (
            objects.vnf_lcm_subscriptions.LccnSubscriptionRequest(
                context=self.context, **fakes.subscription_data))

        result = subscription_obj.create(filter)
        self.assertTrue(filter, result)

    @mock.patch.object(objects.vnf_lcm_subscriptions,
                       '_vnf_lcm_subscriptions_show')
    def test_show(self, mock_vnf_lcm_subscriptions_show):
        filter = fakes.filter
        subscription_obj = \
            objects.vnf_lcm_subscriptions.LccnSubscriptionRequest(
                context=self.context)
        mock_vnf_lcm_subscriptions_show.return_value = \
            '{\
                "filter": "{"operationStates": ["COMPLETED"],\
                "vnfInstanceNames": ["xxxxxxxxxxxxxxxxxx"],\
                "operationTypes": ["INSTANTIATE"],\
                "vnfdIds": ["405d73c7-e964-4c8b-a914-41478ccd7c42"],\
                "vnfProductsFromProviders": [{\
                    "vnfProvider": "x2x", \
                    "vnfProducts": [{\
                        "vnfProductName": "x2xx", \
                        "versions": [{\
                            "vnfSoftwareVersion": "xx2XX", \
                            "vnfdVersions": ["ss2"]\
                        }]\
                    }]\
                }, \
                {\
                    "vnfProvider": "z2z",\
                    "vnfProducts": [{\
                        "vnfProductName": "z2zx", \
                        "versions": [{\
                            "vnfSoftwareVersion": "xx3XX",\
                            "vnfdVersions": \
                            ["s3sx", "s3sa"]\
                        }\
                    ]},\
                    {\
                        "vnfProductName": "zz3ex",\
                        "versions": [{\
                            "vnfSoftwareVersion": "xxe3eXz",\
                            "vnfdVersions": ["ss3xz", "s3esaz"]\
                        },\
                        {\
                            "vnfSoftwareVersion": "xxeeeXw", \
                            "vnfdVersions": ["ss3xw", "ss3w"]\
                        }]\
                    }]\
                }],\
                "notificationTypes": [\
                    "VnfLcmOperationOccurrenceNotification"],\
                "vnfInstanceIds": ["fb0b9a12-4b55-47ac-9ca8-5fdd52c4c07f"]}",\
                "callbackUri": "http://localhost/xxx",\
                "_links": {\
                    "self": {\
                        "href":\
                            "http://localhost:9890//vnflcm/v1/subscriptions\
                            /530a3c43-043a-4b84-9d65-aa0df49f7ced"\
                    }\
                },\
                "id": "530a3c43-043a-4b84-9d65-aa0df49f7ced",\
                "tenant_id": "8cd59fef-397e-4bec-8621-0e69225434f0"\
            }'

        result = subscription_obj.vnf_lcm_subscriptions_show(
            self.context, self.subscription.id)
        self.assertTrue(filter, result)

    @mock.patch.object(objects.vnf_lcm_subscriptions,
                       '_vnf_lcm_subscriptions_all')
    def test_list(self, mock_vnf_lcm_subscriptions_all):
        filter = fakes.filter
        subscription_obj = \
            objects.vnf_lcm_subscriptions.LccnSubscriptionRequest(
                context=self.context)
        mock_vnf_lcm_subscriptions_all.return_value = \
            '{\
                "filter": "{"operationStates": ["COMPLETED"],\
                "vnfInstanceNames": ["xxxxxxxxxxxxxxxxxx"],\
                "operationTypes": ["INSTANTIATE"],\
                "vnfdIds": ["405d73c7-e964-4c8b-a914-41478ccd7c42"],\
                "vnfProductsFromProviders": [{\
                    "vnfProvider": "x2x", \
                    "vnfProducts": [{\
                        "vnfProductName": "x2xx", \
                        "versions": [{\
                            "vnfSoftwareVersion": "xx2XX", \
                            "vnfdVersions": ["ss2"]\
                        }]\
                    }]\
                }, \
                {\
                    "vnfProvider": "z2z",\
                    "vnfProducts": [{\
                        "vnfProductName": "z2zx", \
                        "versions": [{\
                            "vnfSoftwareVersion": "xx3XX",\
                            "vnfdVersions": \
                            ["s3sx", "s3sa"]\
                        }\
                    ]},\
                    {\
                        "vnfProductName": "zz3ex",\
                        "versions": [{\
                            "vnfSoftwareVersion": "xxe3eXz",\
                            "vnfdVersions": ["ss3xz", "s3esaz"]\
                        },\
                        {\
                            "vnfSoftwareVersion": "xxeeeXw", \
                            "vnfdVersions": ["ss3xw", "ss3w"]\
                        }]\
                    }]\
                }],\
                "notificationTypes": [\
                    "VnfLcmOperationOccurrenceNotification"],\
                "vnfInstanceIds": ["fb0b9a12-4b55-47ac-9ca8-5fdd52c4c07f"]}",\
                "callbackUri": "http://localhost/xxx",\
                "_links": {\
                    "self": {\
                        "href":\
                            "http://localhost:9890//vnflcm/v1/subscriptions\
                            /530a3c43-043a-4b84-9d65-aa0df49f7ced"\
                    }\
                },\
                "id": "530a3c43-043a-4b84-9d65-aa0df49f7ced",\
                "tenant_id": "8cd59fef-397e-4bec-8621-0e69225434f0"\
            }'

        result = subscription_obj.vnf_lcm_subscriptions_list(self.context)
        self.assertTrue(filter, result)

    @mock.patch.object(objects.vnf_lcm_subscriptions,
                       '_destroy_vnf_lcm_subscription')
    @mock.patch.object(objects.vnf_lcm_subscriptions, '_get_by_subscriptionid')
    def test_destroy(self, mock_get_by_subscriptionid,
                     mock_vnf_lcm_subscriptions_destroy):
        mock_get_by_subscriptionid.result_value = "OK"
        self.subscription.destroy(self.context, self.subscription.id)
        mock_vnf_lcm_subscriptions_destroy.assert_called_with(
            self.context, self.subscription.id)

    @mock.patch.object(objects.vnf_lcm_subscriptions,
                       '_vnf_lcm_subscriptions_get')
    def test_get(self, mock_vnf_lcm_subscriptions_get):
        subscription_obj = \
            objects.vnf_lcm_subscriptions.LccnSubscriptionRequest(
                context=self.context)
        mock_vnf_lcm_subscriptions_get.return_value = \
            '{\
                "filter": "{"operationStates": ["COMPLETED"],\
                "vnfInstanceNames": ["xxxxxxxxxxxxxxxxxx"],\
                "operationTypes": ["INSTANTIATE"],\
                "vnfdIds": ["405d73c7-e964-4c8b-a914-41478ccd7c42"],\
                "vnfProductsFromProviders": [{\
                    "vnfProvider": "x2x", \
                    "vnfProducts": [{\
                        "vnfProductName": "x2xx", \
                        "versions": [{\
                            "vnfSoftwareVersion": "xx2XX", \
                            "vnfdVersions": ["ss2"]\
                        }]\
                    }]\
                }, \
                {\
                    "vnfProvider": "z2z",\
                    "vnfProducts": [{\
                        "vnfProductName": "z2zx", \
                        "versions": [{\
                            "vnfSoftwareVersion": "xx3XX",\
                            "vnfdVersions": \
                            ["s3sx", "s3sa"]\
                        }\
                    ]},\
                    {\
                        "vnfProductName": "zz3ex",\
                        "versions": [{\
                            "vnfSoftwareVersion": "xxe3eXz",\
                            "vnfdVersions": ["ss3xz", "s3esaz"]\
                        },\
                        {\
                            "vnfSoftwareVersion": "xxeeeXw", \
                            "vnfdVersions": ["ss3xw", "ss3w"]\
                        }]\
                    }]\
                }],\
                "notificationTypes": [\
                    "VnfLcmOperationOccurrenceNotification"],\
                "vnfInstanceIds": ["fb0b9a12-4b55-47ac-9ca8-5fdd52c4c07f"]}",\
                "callbackUri": "http://localhost/xxx",\
                "_links": {\
                    "self": {\
                        "href":\
                            "http://localhost:9890//vnflcm/v1/subscriptions\
                            /530a3c43-043a-4b84-9d65-aa0df49f7ced"\
                    }\
                },\
                "id": "530a3c43-043a-4b84-9d65-aa0df49f7ced",\
                "tenant_id": "8cd59fef-397e-4bec-8621-0e69225434f0"\
            }'

        result = subscription_obj.vnf_lcm_subscriptions_get(self.context,
                'VnfLcmOperationOccurrenceNotification')
        self.assertIn("tenant_id", result)
        result = subscription_obj.vnf_lcm_subscriptions_get(self.context,
                None)
        self.assertIn("tenant_id", result)



# Copyright (c) 2019 OpenStack Foundation
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import routes

from oslo_config import cfg
from tacker.api.vnfpkgm.v1 import controller as vnf_pkgm_controller
from tacker import wsgi


class VnfpkgmAPIRouter(wsgi.Router):
    """Routes requests on the API to the appropriate controller and method."""

    def __init__(self):
        mapper = routes.Mapper()
        super(VnfpkgmAPIRouter, self).__init__(mapper)

    def _setup_route(self, mapper, url, methods, controller, default_resource):
        all_methods = ['HEAD', 'GET', 'POST', 'PUT', 'PATCH', 'DELETE']
        missing_methods = [m for m in all_methods if m not in methods]
        allowed_methods_str = ",".join(methods.keys())
        scope_opts = []
        for method, action in methods.items():
            mapper.connect(url,
                controller=controller,
                action=action,
                conditions={'method': [method]})

            add = cfg.ListOpt('vnfpkgm_'
            + action + '_scope',
                      default=[],
                      help="OAuth2.0 api token scope for" + action)
            scope_opts.append(add)

        cfg.CONF.register_opts(scope_opts, group='authentication')

        if missing_methods:
            mapper.connect(url,
                controller=default_resource,
                action='reject',
                allowed_methods=allowed_methods_str,
                conditions={'method': missing_methods})

    def _setup_routes(self, mapper):
        default_resource = wsgi.Resource(wsgi.DefaultMethodController(),
                                         wsgi.RequestDeserializer())

        controller = vnf_pkgm_controller.create_resource()

        # Allowed methods on /vnf_packages resource
        methods = {"GET": "index", "POST": "create"}
        self._setup_route(mapper, "/vnf_packages",
                methods, controller, default_resource)

        # Allowed methods on /vnf_packages/{id} resource
        methods = {"DELETE": "delete", "GET": "show",
                   "PATCH": "patch"}
        self._setup_route(mapper, "/vnf_packages/{id}",
                methods, controller, default_resource)

        # Allowed methods on /vnf_packages/{id}/package_content resource
        methods = {"PUT": "upload_vnf_package_content",
                   "GET": "fetch_vnf_package_content"}
        self._setup_route(mapper, "/vnf_packages/{id}/package_content",
                methods, controller, default_resource)

        # Allowed methods on
        # /vnf_packages/{id}/package_content/upload_from_uri resource
        methods = {"POST": "upload_vnf_package_from_uri"}
        self._setup_route(mapper,
                "/vnf_packages/{id}/package_content/upload_from_uri",
                methods, controller, default_resource)

        # Allowed methods on /vnf_packages/{id}/vnfd
        methods = {"GET": "get_vnf_package_vnfd"}
        self._setup_route(mapper,
                          "/vnf_packages/{id}/vnfd",
                          methods, controller, default_resource)

        # Allowed methods on /vnf_packages/{id}/artifacts/{artifact_path}
        methods = {"GET": "fetch_vnf_package_artifacts"}
        self._setup_route(mapper,
                          "/vnf_packages/{id}/artifacts/"
                          "{artifact_path:.*?/*.*?}",
                          methods, controller, default_resource)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_serialization import jsonutils as json
from oslo_utils import timeutils
from oslo_versionedobjects import base as ovoo_base
from sqlalchemy.orm import joinedload
from sqlalchemy.sql import text

from tacker.common import exceptions
from tacker.common import utils
from tacker.common.utils import convert_string_to_snakecase
import tacker.conf
from tacker.db import api as db_api
from tacker.db.db_sqlalchemy import api
from tacker.db.db_sqlalchemy import models
from tacker import objects
from tacker.objects import base
from tacker.objects import common
from tacker.objects import fields

_NO_DATA_SENTINEL = object()

LOG = logging.getLogger(__name__)
CONF = tacker.conf.CONF


VNF_INSTANCE_SUBSCRIPTION_FILTER = [
    "vnfdIds", "vnfProvider", "vnfProductName",
    "vnfSoftwareVersion", "vnfdVersions", "vnfInstanceIds",
    "vnfInstanceNames"
]


VNF_INSTANCE_SUBSCRIPTION_FILTER_LISTS = [
    "vnfdIds", "vnfdVersions", "vnfInstanceIds", "vnfInstanceNames"
]


def _get_vnf_subscription_filter_values(vnf_subscription_filter):
    vnfd_ids = vnf_subscription_filter.get('vnfdIds', [])
    vnf_instance_ids = vnf_subscription_filter.get('vnfInstanceIds', [])
    vnf_instance_names = vnf_subscription_filter.get('vnfInstanceNames', [])

    vnfd_products_from_providers = vnf_subscription_filter.get(
        'vnfProductsFromProviders', [])

    vnf_provider = ""
    vnf_product_name = ""
    vnf_software_version = ""
    vnfd_versions = []
    if vnfd_products_from_providers:
        vnfd_products_from_providers = vnfd_products_from_providers[0]
        vnf_provider = vnfd_products_from_providers.get('vnfProvider', "")
        vnf_products = vnfd_products_from_providers.get('vnfProducts', [])

        if vnf_products:
            vnf_product_name = vnf_products[0].get('vnfProductName', "")
            versions = vnf_products[0].get('versions', [])
            if versions:
                vnf_software_version = \
                    versions[0].get('vnfSoftwareVersion', "")
                vnfd_versions = versions[0].get('vnfdVersions', [])

    vnf_subscription_array = [
        {'vnfdIds': vnfd_ids},
        {'vnfInstanceIds': vnf_instance_ids},
        {'vnfInstanceNames': vnf_instance_names},
        {'vnfProvider': vnf_provider},
        {'vnfProductName': vnf_product_name},
        {'vnfSoftwareVersion': vnf_software_version},
        {'vnfdVersions': vnfd_versions}]

    return vnf_subscription_array


def _make_list(value):
    if isinstance(value, list):
        res = ""
        for i in range(len(value)):
            t = "\"{}\"".format(value[i])
            if i == 0:
                res = str(t)
            else:
                res = "{0},{1}".format(res, t)

        res = "[{}]".format(res)
    else:
        res = "[\"{}\"]".format(str(value))
    return res


@db_api.context_manager.reader
def _vnf_lcm_subscriptions_get(context,
                               notification_type,
                               operation_type=None
                               ):

    if notification_type == 'VnfLcmOperationOccurrenceNotification':
        sql = (
            "select"
            " t1.id,t1.callback_uri,t1.authentication,"
            " t1.tenant_id, t2.filter "
            " from "
            " vnf_lcm_subscriptions t1, "
            " (select distinct subscription_uuid,filter from vnf_lcm_filters "
            " where "
            " (notification_types_len = 0 \
                or JSON_CONTAINS(notification_types, '" +
            _make_list(notification_type) +
            "')) "
            " and "
            " (operation_types_len = 0 or JSON_CONTAINS(operation_types, '" +
            _make_list(operation_type) +
            "')) "
            " order by "
            "         notification_types_len desc,"
            "         operation_types_len desc"
            ") t2 "
            " where "
            " t1.id=t2.subscription_uuid "
            " and t1.deleted=0")
    else:
        sql = (
            "select"
            " t1.id,t1.callback_uri,t1.authentication,"
            " t1.tenant_id, t2.filter "
            " from "
            " vnf_lcm_subscriptions t1, "
            " (select distinct subscription_uuid,filter from vnf_lcm_filters "
            " where "
            " (notification_types_len = 0 or \
                JSON_CONTAINS(notification_types, '" +
            _make_list(notification_type) +
            "')) "
            " order by "
            "         notification_types_len desc,"
            "         operation_types_len desc"
            ") t2 "
            " where "
            " t1.id=t2.subscription_uuid "
            " and t1.deleted=0")

    result_list = []
    result = context.session.execute(sql)
    for line in result:
        result_list.append(line)
    return result_list


@db_api.context_manager.reader
def _vnf_lcm_subscriptions_show(context, subscriptionId):

    sql = text(
        "select "
        "t1.id,t1.callback_uri,t1.tenant_id,t2.filter "
        "from vnf_lcm_subscriptions t1, "
        "(select distinct subscription_uuid,filter from vnf_lcm_filters) t2 "
        "where t1.id = t2.subscription_uuid "
        "and deleted = 0 "
        "and t1.id = :subsc_id")
    result_line = ""
    try:
        result = context.session.execute(sql, {'subsc_id': subscriptionId})
        for line in result:
            result_line = line
    except exceptions.NotFound:
        return ''
    except Exception as e:
        raise e
    return result_line


@db_api.context_manager.reader
def _vnf_lcm_subscriptions_all(context):

    sql = text(
        "select "
        "t1.id,t1.callback_uri,t2.filter "
        "from vnf_lcm_subscriptions t1, "
        "(select distinct subscription_uuid,filter from vnf_lcm_filters) t2 "
        "where t1.id = t2.subscription_uuid "
        "and deleted = 0 ")
    result_list = []
    try:
        result = context.session.execute(sql)
        for line in result:
            result_list.append(line)
    except Exception as e:
        raise e

    return result_list


@db_api.context_manager.reader
def _get_by_subscriptionid(context, subscriptionsId):

    sql = text("select id "
             "from vnf_lcm_subscriptions "
             "where id = :subsc_id "
             "and deleted = 0 ")
    try:
        result = context.session.execute(sql, {'subsc_id': subscriptionsId})
    except exceptions.NotFound:
        return ''
    except Exception as e:
        raise e

    return result


@db_api.context_manager.reader
def _vnf_lcm_subscriptions_id_get(context,
                                  callbackUri,
                                  notification_type=None,
                                  operation_type=None,
                                  operation_state=None,
                                  vnf_instance_subscription_filter=None
                                  ):

    sql = ("select "
          "t1.id "
           "from "
           "vnf_lcm_subscriptions t1, "
           "(select subscription_uuid from vnf_lcm_filters "
           "where ")

    if vnf_instance_subscription_filter:
        included_in_filter = []
        column_list = _get_vnf_subscription_filter_values(
            vnf_instance_subscription_filter)

        sql_lst = [sql]
        for column in column_list:
            for key in column:
                if key in VNF_INSTANCE_SUBSCRIPTION_FILTER:
                    value = column[key]
                    if key in VNF_INSTANCE_SUBSCRIPTION_FILTER_LISTS:
                        if value:
                            value = _make_list(value)
                            sql_lst.append(
                                " JSON_CONTAINS({}, '{}') and ".format(
                                    convert_string_to_snakecase(key), value))
                        else:
                            sql_lst.append(" {}_len=0 and ".format(
                                convert_string_to_snakecase(key)))
                    else:
                        sql_lst.append(" {}='{}' and ".format(
                            convert_string_to_snakecase(key), value))

                    included_in_filter.append(key)
        sql = ''.join(sql_lst)

        not_included_in_filter = list(
            set(VNF_INSTANCE_SUBSCRIPTION_FILTER_LISTS) -
            set(included_in_filter))

        # items not being searched for is excluded by adding
        # <name>_len=0 to the sql query
        for key in not_included_in_filter:
            sql = sql + " {}_len=0 and ".format(
                convert_string_to_snakecase(key))

    if notification_type:
        sql = (sql + " JSON_CONTAINS(notification_types, '" +
               _make_list(notification_type) + "') ")
    else:
        sql = sql + " notification_types_len=0 "
    sql = sql + "and "

    if operation_state:
        sql = (sql + " JSON_CONTAINS(operation_states, '" +
               _make_list(operation_state) + "') ")
    else:
        sql = sql + " operation_states_len=0 "
    sql = sql + "and "

    if operation_type:
        sql = sql + " JSON_CONTAINS(operation_types, '" + \
            _make_list(operation_type) + "') "
    else:
        sql = sql + " operation_types_len=0 "
    sql = (
        sql +
        ") t2 where t1.id=t2.subscription_uuid and t1.callback_uri= '" +
        callbackUri +
        "' and t1.deleted=0 ")
    LOG.debug("sql[%s]" % sql)

    try:
        result = context.session.execute(sql)
        for line in result:
            return line
    except exceptions.NotFound:
        return ''
    except Exception as exc:
        LOG.error("SQL Error: %s" % str(exc))
        return ''


def _add_filter_data(context, subscription_id, filter):
    with db_api.context_manager.writer.using(context):

        vnf_products_from_providers = (
            filter.get('vnfInstanceSubscriptionFilter', {})
            .get('vnfProductsFromProviders'))

        if vnf_products_from_providers:
            vnf_products_from_providers = vnf_products_from_providers[0]

        new_entries = []
        new_entries.append({"subscription_uuid": subscription_id,
                            "filter": filter,
                            "vnf_products_from_providers":
                                vnf_products_from_providers})

        context.session.execute(
            models.VnfLcmFilters.__table__.insert(None),
            new_entries)


@db_api.context_manager.reader
def _vnf_lcm_subscription_list_by_filters(context,
        read_deleted=None, filters=None, nextpage_opaque_marker=None):
    query = api.model_query(context, models.VnfLcmSubscriptions,
                            read_deleted=read_deleted,
                            project_only=True)
    binary_columns = ['notification_types', 'operation_types']

    if filters:
        filter_data = json.dumps(filters)
        if 'ChangeNotificationsFilter' in filter_data:
            query = query.join(models.VnfLcmFilters)

        if 'and' in filters:
            filters_and = []
            for filter in filters['and']:
                if filter['field'] in binary_columns:
                    converted_value = utils.str_to_bytes(filter['value'])
                    filter['value'] = converted_value
                filters_and.append(filter)

            filters = {'and': filters_and}
        else:
            if filters['field'] in binary_columns:
                converted_value = utils.str_to_bytes(filters['value'])
                filters.update({'value': converted_value})

        query = common.apply_filters(query, filters)

    if nextpage_opaque_marker:
        start_offset = CONF.vnf_lcm.subscription_num * nextpage_opaque_marker
        return query.order_by(
            models.VnfLcmSubscriptions.created_at).limit(
                CONF.vnf_lcm.subscription_num + 1).offset(
                    start_offset).all()
    else:
        return query.order_by(models.VnfLcmSubscriptions.created_at).all()


@db_api.context_manager.writer
def _vnf_lcm_subscriptions_create(context, values, filter):
    with db_api.context_manager.writer.using(context):

        new_entries = []
        if 'authentication' in values:
            new_entries.append({"id": values.id,
                                "callback_uri": values.callback_uri,
                                "authentication":
                                    values.authentication,
                                "tenant_id": values.tenant_id})
        else:
            new_entries.append({"id": values.id,
                                "callback_uri": values.callback_uri,
                                "tenant_id": values.tenant_id})

        context.session.execute(
            models.VnfLcmSubscriptions.__table__.insert(None),
            new_entries)

        callbackUri = values.callback_uri
        if filter:
            notification_type = filter.get('notificationTypes')
            operation_type = filter.get('operationTypes')
            operation_state = filter.get('operationStates')
            subscription_filter = filter.get('vnfInstanceSubscriptionFilter')

            vnf_lcm_subscriptions_id = _vnf_lcm_subscriptions_id_get(
                context,
                callbackUri,
                notification_type=notification_type,
                operation_type=operation_type,
                operation_state=operation_state,
                vnf_instance_subscription_filter=subscription_filter)

            if vnf_lcm_subscriptions_id:
                raise exceptions.SeeOther(message=vnf_lcm_subscriptions_id.id)

            _add_filter_data(context, values.id, filter)

        else:
            vnf_lcm_subscriptions_id = _vnf_lcm_subscriptions_id_get(context,
                                            callbackUri)

            if vnf_lcm_subscriptions_id:
                raise exceptions.SeeOther(message=vnf_lcm_subscriptions_id.id)
            _add_filter_data(context, values.id, {})

    return values


@db_api.context_manager.writer
def _destroy_vnf_lcm_subscription(context, subscriptionId):
    now = timeutils.utcnow()
    updated_values = {'deleted': 1,
                      'deleted_at': now}
    try:
        api.model_query(context, models.VnfLcmSubscriptions). \
            filter_by(id=subscriptionId). \
            update(updated_values, synchronize_session=False)
    except Exception as e:
        raise e


@db_api.context_manager.reader
def _subscription_get_by_id(context, subscription_uuid, columns_to_join=None):

    query = api.model_query(context, models.VnfLcmSubscriptions,
                            read_deleted="no", project_only=True). \
        filter_by(id=subscription_uuid)

    if columns_to_join:
        for column in columns_to_join:
            query = query.options(joinedload(column))

    result = query.first()

    if not result:
        raise exceptions.NotFound(resource='Subscription',
            id=subscription_uuid)

    return result


def _make_subscription_list(context, subscription_list, db_subscription_list,
                            expected_attrs=None):
    subscription_cls = LccnSubscription

    subscription_list.objects = []
    for db_subscription in db_subscription_list:
        subscription_obj = subscription_cls._from_db_object(
            context, subscription_cls(context), db_subscription,
            expected_attrs=expected_attrs)
        subscription_list.objects.append(subscription_obj)

    subscription_list.obj_reset_changes()
    return subscription_list


@base.TackerObjectRegistry.register
class LccnSubscriptionRequest(base.TackerObject, base.TackerPersistentObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(nullable=False),
        'callback_uri': fields.StringField(nullable=False),
        # TODO(YiFeng) define SubscriptionAuthentication object
        'authentication':
            fields.StringField(nullable=True),
        'filter': fields.StringField(nullable=True),
        'tenant_id': fields.StringField(nullable=False)
    }

    @base.remotable
    def create(self, filter):
        updates = self.obj_clone()
        db_vnf_lcm_subscriptions = _vnf_lcm_subscriptions_create(
            self._context, updates, filter)

        LOG.debug(
            'test_log: db_vnf_lcm_subscriptions %s' %
            db_vnf_lcm_subscriptions)

        return db_vnf_lcm_subscriptions

    @base.remotable_classmethod
    def vnf_lcm_subscriptions_show(cls, context, subscriptionId):
        try:
            vnf_lcm_subscriptions = _vnf_lcm_subscriptions_show(
                context, subscriptionId)
        except Exception as e:
            raise e
        return vnf_lcm_subscriptions

    @base.remotable_classmethod
    def vnf_lcm_subscriptions_list(cls, context):
        # get vnf_lcm_subscriptions data
        try:
            vnf_lcm_subscriptions = _vnf_lcm_subscriptions_all(context)
        except Exception as e:
            raise e

        return vnf_lcm_subscriptions

    @base.remotable_classmethod
    def vnf_lcm_subscriptions_get(cls, context,
                                  notification_type,
                                  operation_type=None):
        return _vnf_lcm_subscriptions_get(context,
                                          notification_type,
                                          operation_type)

    @base.remotable_classmethod
    def destroy(cls, context, subscriptionId):
        try:
            get_subscriptionid = _get_by_subscriptionid(
                context, subscriptionId)
        except Exception as e:
            raise e

        if not get_subscriptionid:
            return 404

        try:
            _destroy_vnf_lcm_subscription(context, subscriptionId)
        except Exception as e:
            raise e

        return 204


@base.TackerObjectRegistry.register
class ChangeNotificationsFilter(
        base.TackerObject, base.TackerPersistentObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(nullable=False),
        'subscription_uuid': fields.UUIDField(nullable=False),
        'filter': fields.StringField(nullable=True),
        'vnf_products_from_providers':
            fields.StringField(nullable=True),
        'vnfd_ids': fields.StringField(nullable=True),
        'vnfd_ids_len': fields.IntegerField(
            nullable=True, default=0),
        'vnf_provider': fields.StringField(nullable=True),
        'vnf_product_name': fields.StringField(nullable=True),
        'vnf_software_version': fields.StringField(nullable=True),
        'vnfd_versions': fields.StringField(nullable=True),
        'vnfd_versions_len': fields.IntegerField(
            nullable=True, default=0),
        'vnf_instance_ids': fields.StringField(nullable=True),
        'vnf_instance_ids_len': fields.IntegerField(
            nullable=True, default=0),
        'vnf_instance_names': fields.StringField(nullable=True),
        'vnf_instance_names_len': fields.IntegerField(
            nullable=True, default=0),
        'notification_types': fields.StringField(nullable=True),
        'notification_types_len': fields.IntegerField(
            nullable=True, default=0),
        'operation_types': fields.StringField(nullable=True),
        'operation_types_len': fields.IntegerField(
            nullable=True, default=0),
        'operation_states': fields.StringField(nullable=True),
        'operation_states_len': fields.IntegerField(
            nullable=True, default=0),
        'tenant_id': fields.StringField(nullable=False),
    }


@base.TackerObjectRegistry.register
class ChangeNotificationsFilterList(
        ovoo_base.ObjectListBase, base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('ChangeNotificationsFilter')
    }


@base.TackerObjectRegistry.register
class LccnSubscription(base.TackerObject, base.TackerPersistentObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(nullable=False),
        'callback_uri': fields.StringField(nullable=False),
        'filter': fields.ObjectField(
            'ChangeNotificationsFilter', nullable=True),
    }

    ALL_ATTRIBUTES = {
        'id': ('id', 'uuid',
            'VnfLcmSubscriptions'),
        'vnfdIds': ('vnfd_ids', 'string',
            'VnfLcmFilters'),
        'vnfProvider': ('vnf_provider', 'string',
            'VnfLcmFilters'),
        'vnfProductName': ('vnf_product_name', 'string',
            'VnfLcmFilters'),
        'vnfSoftwareVersion': ('vnf_software_version', 'string',
            'VnfLcmFilters'),
        'vnfdVersions': ('vnfd_versions', 'string',
            'VnfLcmFilters'),
        'vnfInstanceIds': ('vnf_instance_ids', 'string',
            'VnfLcmFilters'),
        'vnfInstanceNames': ('vnf_instance_names', 'string',
            'VnfLcmFilters'),
        'notificationTypes': ('notification_types', 'string',
            'VnfLcmFilters'),
        'operationTypes': ('operation_types', 'string',
            'VnfLcmFilters'),
        'operationStates': ('operation_states', 'string',
            'VnfLcmFilters'),
        'callbackUri': ('callback_uri', 'string',
            'VnfLcmSubscriptions'),
        'tenantId': ('tenant_id', 'string',
            'VnfLcmSubscriptions'),
    }

    FLATTEN_ATTRIBUTES = utils.flatten_dict(ALL_ATTRIBUTES.copy())

    @staticmethod
    def _from_db_object(context, subscription, db_subscription,
                        expected_attrs=None):
        expected_attrs = expected_attrs or ['filter']

        subscription._context = context

        for key in subscription.fields:
            if key in ['filter']:
                continue
            db_key = key
            setattr(subscription, key, db_subscription[db_key])

        subscription._context = context
        subscription._extra_attributes_from_db_object(
            subscription, db_subscription, expected_attrs)

        subscription.obj_reset_changes()
        return subscription

    @staticmethod
    def _extra_attributes_from_db_object(subscription, db_subscription,
                                         expected_attrs=None):
        """Method to help with migration of extra attributes to objects."""

        if expected_attrs is None:
            expected_attrs = ['filter']

        if 'filter' in expected_attrs:
            subscription._load_subscription_filter(
                db_subscription.get('filter'))

    def _load_subscription_filter(self, db_filter=_NO_DATA_SENTINEL):
        if db_filter is _NO_DATA_SENTINEL:
            subscription = self.get_by_id(
                self._context, self.id,
                expected_attrs=['filter'])
            if 'filter' in subscription:
                self.filter = \
                    subscription.filter
                self.filter.obj_reset_changes(recursive=True)
                self.obj_reset_changes(['filter'])
            else:
                self.filter = \
                    objects.ChangeNotificationsFilterList(objects=[])
        elif db_filter:
            self.filter = base.obj_make_list(
                self._context, objects.ChangeNotificationsFilterList(
                    self._context), objects.ChangeNotificationsFilter,
                db_filter)
            self.obj_reset_changes(['filter'])

    @base.remotable_classmethod
    def get_by_id(cls, context, id, expected_attrs=None):
        db_subscription = _subscription_get_by_id(
            context, id, columns_to_join=expected_attrs)
        return cls._from_db_object(context, cls(), db_subscription,
                                   expected_attrs=expected_attrs)


@base.TackerObjectRegistry.register
class LccnSubscriptionList(ovoo_base.ObjectListBase, base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('LccnSubscription')
    }

    @base.remotable_classmethod
    def get_by_filters(cls, context, read_deleted=None,
                       filters=None, nextpage_opaque_marker=None):

        db_subscriptions = _vnf_lcm_subscription_list_by_filters(context,
                                read_deleted=read_deleted,
                                filters=filters,
                                nextpage_opaque_marker=nextpage_opaque_marker)
        return _make_subscription_list(context, cls(), db_subscriptions)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os

from tacker.tests.functional import base
from tacker.tests.functional.common.fake_server import FakeServerManager
from tacker.tests.functional.sol.vnflcm import base as vnflcm_base


FAKE_SERVER_MANAGER_T1 = FakeServerManager()
FAKE_SERVER_PORT_T1 = 9995
FAKE_SERVER_MANAGER_T2 = FakeServerManager()
FAKE_SERVER_PORT_T2 = 9996


class BaseVnfLcmMultiTenantTest(vnflcm_base.BaseVnfLcmTest):

    prepare_fake_server = False

    @classmethod
    def setUpClass(cls):
        super(BaseVnfLcmMultiTenantTest, cls).setUpClass()

        result = cls.get_openstack_client_session(
            vim_conf_file='local-tenant1-vim.yaml')
        cls.client_tenant1 = result.get('client')
        cls.http_client_tenant1 = result.get('http_client')
        cls.h_client_tenant1 = result.get('h_client')
        cls.glance_client_tenant1 = result.get('glance_client')

        result = cls.get_openstack_client_session(
            vim_conf_file='local-tenant2-vim.yaml')
        cls.client_tenant2 = result.get('client')
        cls.http_client_tenant2 = result.get('http_client')
        cls.h_client_tenant2 = result.get('h_client')
        cls.glance_client_tenant2 = result.get('glance_client')

        cls.tacker_client_t1 = base.BaseTackerTest.tacker_http_client(
            'local-tenant1-vim.yaml')
        cls.tacker_client_t2 = base.BaseTackerTest.tacker_http_client(
            'local-tenant2-vim.yaml')

        # Set up fake NFVO server for tenant1 and tenant2
        cls.servers = {FAKE_SERVER_PORT_T1: FAKE_SERVER_MANAGER_T1,
                       FAKE_SERVER_PORT_T2: FAKE_SERVER_MANAGER_T2}
        # NOTE: Create both server in parallel, otherwise they can
        # cause (especially server start) job timeout.
        for port, manager in cls.servers.items():
            cls._prepare_start_fake_server(manager, port)

    @classmethod
    def tearDownClass(cls):
        super(BaseVnfLcmMultiTenantTest, cls).tearDownClass()
        for _, manager in cls.servers.items():
            manager.stop_server()

    def setUp(self):
        super(BaseVnfLcmMultiTenantTest, self).setUp()
        self.base_url = "/vnfpkgm/v1/vnf_packages"

        callback_url = os.path.join(
            vnflcm_base.MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        self._clear_history_and_set_callback(FAKE_SERVER_MANAGER_T1,
            callback_url)
        self._clear_history_and_set_callback(FAKE_SERVER_MANAGER_T2,
            callback_url)

        vim_list = self.client.list_vims()
        self.vim_tenant1 = self.get_vim(vim_list, 'VIM_TEST')
        if not self.vim_tenant1:
            assert False, "vim_list is Empty: Tenant VIM is missing"
        self.vim_tenant2 = self.get_vim(vim_list, 'VIM_DEMO')
        if not self.vim_tenant2:
            assert False, "vim_list is Empty: Tenant VIM is missing"

        result = self._create_network_settings()
        self.ext_networks_tenant1 = result.get('ext_networks')
        self.ext_vl_tenant1 = result.get('ext_vl')
        self.ext_mngd_networks_tenant1 = result.get('ext_mngd_networks')
        self.ext_link_ports_tenant1 = result.get('ext_link_ports')
        self.ext_subnets_tenant1 = result.get('ext_subnets')
        self.changed_ext_networks_tenant1 = result.get(
            'changed_ext_networks')
        self.changed_ext_subnets_tenant1 = result.get(
            'changed_ext_subnets')

        result = self._create_network_settings()
        self.ext_networks_tenant2 = result.get('ext_networks')
        self.ext_vl_tenant2 = result.get('ext_vl')
        self.ext_mngd_networks_tenant2 = result.get('ext_mngd_networks')
        self.ext_link_ports_tenant2 = result.get('ext_link_ports')
        self.ext_subnets_tenant2 = result.get('ext_subnets')
        self.changed_ext_networks_tenant2 = result.get(
            'changed_ext_networks')
        self.changed_ext_subnets_tenant2 = result.get(
            'changed_ext_subnets')

    @classmethod
    def get_openstack_client_session(cls, vim_conf_file):
        client = base.BaseTackerTest.tackerclient(vim_conf_file)
        http_client = base.BaseTackerTest.tacker_http_client(vim_conf_file)
        h_client = base.BaseTackerTest.heatclient(vim_conf_file)
        glance_client = base.BaseTackerTest.glanceclient(vim_conf_file)
        return {'client': client,
                'http_client': http_client,
                'h_client': h_client,
                'glance_client': glance_client}



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils import uuidutils

from tacker import context
from tacker.nfvo.drivers.vim import openstack_driver
from tacker.tests.unit import base
from tacker.tests.unit.db import utils


class FakeNeutronClient(mock.Mock):
    def __init__(self):
        super(FakeNeutronClient, self).__init__()
        self.__fc_dict = {}
        self.__pp_dict = {}
        self.__ppg_dict = {}
        self.__chain_dict = {}

    def flow_classifier_create(self, fc_create_dict):
        fc_id = uuidutils.generate_uuid()
        self.__fc_dict[fc_id] = fc_create_dict
        return fc_id

    def show_flow_classifier(self, fc_dict):
        fc_name = fc_dict['name']
        for fc_id in self.__fc_dict:
            fc = self.__fc_dict[fc_id]
            if fc_name == fc['name']:
                return {'id': fc_id}

        return None

    def flow_classifier_update(self, fc_id, fc_update_dict):
        if fc_id not in self.__fc_dict:
            return None
        self.__fc_dict[fc_id] = fc_update_dict
        return fc_update_dict

    def flow_classifier_delete(self, fc_id):
        if fc_id not in self.__fc_dict:
            raise ValueError('fc not found')
        self.__fc_dict.pop(fc_id)

    def port_pair_create(self, port_pair):
        pp_id = uuidutils.generate_uuid()
        self.__pp_dict[pp_id] = port_pair
        return pp_id

    def port_pair_list(self):
        pp = {'port_pairs': [{'ingress': 'abc',
                              'egress': 'xyz'}]}
        return pp

    def show_port_pair(self, port_pair_dict):
        input_pp_name = port_pair_dict['name']
        for pp_id in self.__pp_dict:
            port_pair = self.__pp_dict[pp_id]
            if port_pair['name'] == input_pp_name:
                return {'id': pp_id}

        return None

    def port_pair_group_create(self, port_pair_group):
        ppg_id = uuidutils.generate_uuid()
        self.__ppg_dict[ppg_id] = port_pair_group
        return ppg_id

    def port_pair_group_list(self):
        value = {'port_pair_groups': []}
        return value

    def show_port_pair_group(self, port_pair_group_dict):
        input_ppg_name = port_pair_group_dict['name']
        for ppg_id in self.__ppg_dict:
            port_pair_group = self.__ppg_dict[ppg_id]
            if port_pair_group['name'] == input_ppg_name:
                return {'id': ppg_id}

        return None

    def port_chain_create(self, port_chain):
        chain_id = uuidutils.generate_uuid()
        self.__chain_dict[chain_id] = port_chain
        return chain_id

    def show_port_chain(self, port_chain_dict):
        input_chain_name = port_chain_dict['name']
        for chain_id in self.__chain_dict:
            port_chain = self.__chain_dict[chain_id]
            if port_chain['name'] == input_chain_name:
                return {'id': chain_id}
        return None

    def port_chain_delete(self, chain_id):
        if chain_id not in self.__chain_dict:
            raise ValueError('port chain delete failed')
        self.__chain_dict.pop(chain_id)

    def port_chain_list(self):
        pc = {'port_chains': [{'port_pair_groups': ['random_id1',
                                                    'random_id2']}]}
        return pc


class TestChainSFC(base.TestCase):

    def setUp(self):
        super(TestChainSFC, self).setUp()
        self.context = context.get_admin_context()
        self.sfc_driver = openstack_driver.OpenStack_Driver()
        self._mock_neutron_client()
        self.addCleanup(mock.patch.stopall)

    def _mock_neutron_client(self):
        self.neutron_client = mock.Mock(wraps=FakeNeutronClient())
        fake_neutron_client = mock.Mock()
        fake_neutron_client.return_value = self.neutron_client
        self._mock(
            'tacker.nfvo.drivers.vim.openstack_driver.'
            'NeutronClient',
            fake_neutron_client)

    def _mock(self, target, new=mock.DEFAULT):
        patcher = mock.patch(target, new)
        return patcher.start()

    def test_create_flow_classifier(self):
        flow_classifier = {'name': 'fake_fc',
                           'source_port_range': '2005-2010',
                           'ip_proto': 6,
                           'destination_port_range': '80-180'}
        result = self.sfc_driver.\
            create_flow_classifier(name='fake_ffg', fc=flow_classifier,
                                   auth_attr=utils.get_vim_auth_obj())
        self.assertIsNotNone(result)

    def test_update_flow_classifier(self):
        auth_attr = utils.get_vim_auth_obj()
        flow_classifier = {'name': 'next_fake_fc',
                           'source_port_range': '2005-2010',
                           'ip_proto': 6,
                           'destination_port_range': '80-180'}

        fc_id = self.sfc_driver.\
            create_flow_classifier(name='fake_ffg', fc=flow_classifier,
                                   auth_attr=utils.get_vim_auth_obj())

        flow_classifier_update = {'name': 'next_fake_fc_two',
                                  'instance_id': None,
                                  'status': 'PENDING_CREATE',
                                  'match': {'source_port_range': '5-10',
                                            'ip_proto': 17,
                                            'destination_port_range': '2-4'}}

        fc_id = self.sfc_driver.\
            create_flow_classifier(name='fake_ffg', fc=flow_classifier,
                                   auth_attr=utils.get_vim_auth_obj())

        self.assertIsNotNone(fc_id)

        vnf_1 = {'name': 'test_create_chain_vnf_1',
                 'connection_points': [uuidutils.generate_uuid(),
                                       uuidutils.generate_uuid()]}
        vnf_2 = {'name': 'test_create_chain_vnf_2',
                 'connection_points': [uuidutils.generate_uuid(),
                                       uuidutils.generate_uuid()]}
        vnf_3 = {'name': 'test_create_chain_vnf_3',
                 'connection_points': [uuidutils.generate_uuid(),
                                       uuidutils.generate_uuid()]}
        vnfs = [vnf_1, vnf_2, vnf_3]

        chain_id = self.sfc_driver.create_chain(name='fake_ffg',
                                                path_id=None,
                                                fc_ids=fc_id,
                                                vnfs=vnfs,
                                                auth_attr=auth_attr)
        self.assertIsNotNone(chain_id)

        result = self.sfc_driver.\
            update_flow_classifier(chain_id=chain_id,
                                   fc=flow_classifier_update,
                                   auth_attr=utils.get_vim_auth_obj())
        self.assertIsNotNone(result)

    def test_delete_flow_classifier(self):
        flow_classifier = {'name': 'another_fake_fc',
                           'description': 'another flow-classifier',
                           'source_port_range': '1999-2005',
                           'ip_proto': 6,
                           'destination_port_range': '80-100'}
        fc_id = self.sfc_driver.\
            create_flow_classifier(name='fake_ffg', fc=flow_classifier,
                                   auth_attr=utils.get_vim_auth_obj())

        self.assertIsNotNone(fc_id)

        try:
            self.sfc_driver.\
                delete_flow_classifier(fc_id=fc_id,
                                       auth_attr=utils.get_vim_auth_obj())
        except Exception:
            self.assertTrue(True)

    def test_create_chain(self):
        auth_attr = utils.get_vim_auth_obj()
        flow_classifier = {'name': 'test_create_chain_fc',
                           'description': 'fc for testing create chain',
                           'source_port_range': '1997-2008',
                           'ip_proto': 6,
                           'destination_port_range': '80-100'}
        fc_id = self.sfc_driver.\
            create_flow_classifier(name='fake_ffg', fc=flow_classifier,
                                   auth_attr=auth_attr)

        self.assertIsNotNone(fc_id)

        vnf_1 = {'name': 'test_create_chain_vnf_1',
                 'connection_points': [uuidutils.generate_uuid(),
                                       uuidutils.generate_uuid()]}
        vnf_2 = {'name': 'test_create_chain_vnf_2',
                 'connection_points': [uuidutils.generate_uuid(),
                                       uuidutils.generate_uuid()]}
        vnf_3 = {'name': 'test_create_chain_vnf_3',
                 'connection_points': [uuidutils.generate_uuid(),
                                       uuidutils.generate_uuid()]}
        vnfs = [vnf_1, vnf_2, vnf_3]

        result = self.sfc_driver.create_chain(name='fake_ffg',
                                              path_id=None,
                                              fc_ids=fc_id,
                                              vnfs=vnfs,
                                              auth_attr=auth_attr)

        self.assertIsNotNone(result)

    def test_delete_chain(self):
        auth_attr = utils.get_vim_auth_obj()
        flow_classifier = {'name': 'test_delete_chain_fc',
                           'description': 'fc for testing delete chain',
                           'source_port_range': '1000-2000',
                           'ip_proto': 6,
                           'destination_port_range': '80-180'}
        fc_id = self.sfc_driver.\
            create_flow_classifier(name='fake_ffg', fc=flow_classifier,
                                   auth_attr=auth_attr)

        self.assertIsNotNone(fc_id)

        vnf_1 = {'name': 'test_delete_chain_vnf_1',
                 'connection_points': [uuidutils.generate_uuid(),
                                       uuidutils.generate_uuid()]}
        vnf_2 = {'name': 'test_delete_chain_vnf_2',
                 'connection_points': [uuidutils.generate_uuid(),
                                       uuidutils.generate_uuid()]}
        vnf_3 = {'name': 'test_delete_chain_vnf_3',
                 'connection_points': [uuidutils.generate_uuid(),
                                       uuidutils.generate_uuid()]}
        vnfs = [vnf_1, vnf_2, vnf_3]

        chain_id = self.sfc_driver.create_chain(name='fake_ffg',
                                                path_id=None,
                                                fc_ids=fc_id,
                                                vnfs=vnfs,
                                                auth_attr=auth_attr)

        self.assertIsNotNone(chain_id)

        try:
            self.sfc_driver.delete_chain(chain_id,
                                         auth_attr=auth_attr)
        except Exception:
            self.assertTrue(True)



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from tacker.sol_refactored.conductor import conductor_rpc_v2
from tacker.sol_refactored.conductor import conductor_v2
from tacker.sol_refactored.objects import base as objects_base


class ConductorV2Hook(object):

    def initialize_service_hook(self, service):
        endpoints = [conductor_v2.ConductorV2()]
        serializer = objects_base.TackerObjectSerializer()
        service.conn.create_consumer(
            conductor_rpc_v2.TOPIC_CONDUCTOR_V2, endpoints,
            serializer=serializer)



# Copyright 2013 Cisco Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Common utilities and helper functions for Openstack Networking Plugins.
"""

from tacker._i18n import _
from tacker.common import exceptions as n_exc
from tacker.common import utils
from tacker.plugins.common import constants


def verify_vlan_range(vlan_range):
    """Raise an exception for invalid tags or malformed range."""
    for vlan_tag in vlan_range:
        if not utils.is_valid_vlan_tag(vlan_tag):
            raise n_exc.NetworkVlanRangeError(
                vlan_range=vlan_range,
                error=_("%s is not a valid VLAN tag") % vlan_tag)
    if vlan_range[1] < vlan_range[0]:
        raise n_exc.NetworkVlanRangeError(
            vlan_range=vlan_range,
            error=_("End of VLAN range is less than start of VLAN range"))


def parse_network_vlan_range(network_vlan_range):
    """Interpret a string as network[:vlan_begin:vlan_end]."""
    entry = network_vlan_range.strip()
    if ':' in entry:
        try:
            network, vlan_min, vlan_max = entry.split(':')
            vlan_range = (int(vlan_min), int(vlan_max))
        except ValueError as ex:
            raise n_exc.NetworkVlanRangeError(vlan_range=entry, error=ex)
        verify_vlan_range(vlan_range)
        return network, vlan_range
    else:
        return entry, None


def parse_network_vlan_ranges(network_vlan_ranges_cfg_entries):
    """Interpret a list of strings as network[:vlan_begin:vlan_end] entries."""
    networks = {}
    for entry in network_vlan_ranges_cfg_entries:
        network, vlan_range = parse_network_vlan_range(entry)
        if vlan_range:
            networks.setdefault(network, []).append(vlan_range)
        else:
            networks.setdefault(network, [])
    return networks


def in_pending_status(status):
    return status in (constants.PENDING_CREATE,
                      constants.PENDING_UPDATE,
                      constants.PENDING_DELETE)



# Copyright 2016 Brocade Communications System, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaml

from oslo_config import cfg
from tackerclient.common import exceptions

from tacker.plugins.common import constants as evt_constants
from tacker.tests import constants
from tacker.tests.functional import base
from tacker.tests.utils import read_file

import time
CONF = cfg.CONF


class NsdTestCreate(base.BaseTackerTest):

    def _test_create_tosca_vnfd(self, tosca_vnfd_file, vnfd_name):
        input_yaml = read_file(tosca_vnfd_file)
        tosca_dict = yaml.safe_load(input_yaml)
        tosca_arg = {'vnfd': {'name': vnfd_name,
                              'attributes': {'vnfd': tosca_dict}}}
        vnfd_instance = self.client.create_vnfd(body=tosca_arg)
        self.assertEqual(vnfd_instance['vnfd']['name'], vnfd_name)
        self.assertIsNotNone(vnfd_instance)

        vnfds = self.client.list_vnfds().get('vnfds')
        self.assertIsNotNone(vnfds, "List of vnfds are Empty after Creation")
        return vnfd_instance['vnfd']['id']

    def _test_create_nsd(self, tosca_nsd_file, nsd_name):
        input_yaml = read_file(tosca_nsd_file)
        tosca_dict = yaml.safe_load(input_yaml)
        tosca_arg = {'nsd': {'name': nsd_name,
                             'attributes': {'nsd': tosca_dict}}}
        nsd_instance = self.client.create_nsd(body=tosca_arg)
        self.assertIsNotNone(nsd_instance)
        return nsd_instance['nsd']['id']

    def _test_delete_nsd(self, nsd_id):
        try:
            self.client.delete_nsd(nsd_id)
        except Exception:
            assert False, "nsd Delete failed"

    def _test_delete_vnfd(self, vnfd_id, timeout=constants.NS_DELETE_TIMEOUT):
        start_time = int(time.time())
        while True:
            try:
                self.client.delete_vnfd(vnfd_id)
            except exceptions.Conflict:
                time.sleep(2)
            except Exception:
                assert False, "vnfd Delete failed"
            else:
                break
            if (int(time.time()) - start_time) > timeout:
                assert False, "vnfd still in use"
        self.verify_vnfd_events(vnfd_id, evt_constants.RES_EVT_DELETE,
                                evt_constants.RES_EVT_NA_STATE)

    def test_create_delete_nsd(self):
        vnfd1_id = self._test_create_tosca_vnfd(
            'test-nsd-vnfd1.yaml',
            'test-nsd-vnfd1')
        vnfd2_id = self._test_create_tosca_vnfd(
            'test-nsd-vnfd2.yaml',
            'test-nsd-vnfd2')
        nsd_id = self._test_create_nsd(
            'test-nsd.yaml',
            'test-nsd')
        self._test_delete_nsd(nsd_id)
        self._test_delete_vnfd(vnfd1_id)
        self._test_delete_vnfd(vnfd2_id)



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker import context
from tacker.sol_refactored.common import lcm_op_occ_utils as lcmocc_utils
from tacker.sol_refactored import objects
from tacker.sol_refactored.objects.v2 import fields
from tacker.tests import base


# instantiatedVnfInfo examples
# NOTE:
# - some identifiers are modified to make check easy.
# - some attributes which are not related to test update_lcmocc are omitted.
_inst_info_example_1 = {
    # "flavourId", "vnfState", "scaleStatus", "maxScaleLevels" are omitted
    # "extCpInfo": omitted
    "extVirtualLinkInfo": [
        {
            "id": "bbf0932a-6142-4ea8-93cd-8059dba594a1",
            "resourceHandle": {
                "resourceId": "3529d333-dbcc-4d93-9b64-210647712569"
            },
            "extLinkPorts": [
                {
                    "id": "res_id_VDU2_CP1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU2_CP1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU2_CP1"
                },
                {
                    "id": "res_id_VDU1_CP1_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP1_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU1_CP1_1"
                }
            ],
            # "currentVnfExtCpData": omitted
        },
        {
            "id": "790949df-c7b3-4926-a559-3895412f1dfe",
            "resourceHandle": {
                "resourceId": "367e5b3b-34dc-47f2-85b8-c39e3272893a"
            },
            "extLinkPorts": [
                {
                    "id": "res_id_VDU2_CP2",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU2_CP2",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU2_CP2"
                },
                {
                    "id": "res_id_VDU1_CP2_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP2_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU1_CP2_1"
                }
            ],
            # "currentVnfExtCpData": omitted
        },
        # following two data are for test_update_lcmocc_change_ext_conn
        {
            "id": "cba2e572-2cc5-4d11-af60-728615883206",
            "resourceHandle": {
                "resourceId": "4529d333-dbcc-4d93-9b64-210647712569"
            },
            "extLinkPorts": [],
            "currentVnfExtCpData": [
                {
                    "cpdId": "VDU1_CP1",
                    "cpConfig": {
                        "VDU1_CP1_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "numDynamicAddresses": 1
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                },
                {
                    "cpdId": "VDU2_CP1",
                    "cpConfig": {
                        "VDU2_CP1_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "fixedAddresses": [
                                                    "10.10.0.102"
                                                ]
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                }
            ]
        },
        {
            "id": "fc131806-8a4f-43a3-82c0-a38e87fe87bd",
            "resourceHandle": {
                "resourceId": "5529d333-dbcc-4d93-9b64-210647712569"
            },
            "extLinkPorts": [],
            "currentVnfExtCpData": [
                {
                    "cpdId": "VDU1_CP2",
                    "cpConfig": {
                        "VDU1_CP2_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "numDynamicAddresses": 1,
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                },
                {
                    "cpdId": "VDU2_CP2",
                    "cpConfig": {
                        "VDU2_CP2_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "fixedAddresses": [
                                                    "10.10.1.102"
                                                ],
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                }
            ]
        }
    ],
    "extManagedVirtualLinkInfo": [
        {
            "id": "res_id_internalVL1",
            "vnfVirtualLinkDescId": "internalVL1",
            "networkResource": {
                "resourceId": "res_id_internalVL1"
            },
            "vnfLinkPorts": [
                {
                    "id": "res_id_VDU2_CP3",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU2_CP3",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU2_CP3-res_id_VDU2",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "res_id_VDU1_CP3_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP3_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU1_CP3-res_id_VDU1_1",
                    "cpInstanceType": "VNFC_CP"
                }
            ]
        }
    ],
    "vnfcResourceInfo": [
        {
            "id": "res_id_VDU1_1",
            "vduId": "VDU1",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU1_1",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "storageResourceIds": [
                "res_id_VirtualStorage_1"
            ],
            "vnfcCpInfo": [
                {
                    "id": "VDU1_CP1-res_id_VDU1_1",
                    "cpdId": "VDU1_CP1",
                    "vnfExtCpId": "cp-res_id_VDU1_CP1_1"
                },
                {
                    "id": "VDU1_CP2-res_id_VDU1_1",
                    "cpdId": "VDU1_CP2",
                    "vnfExtCpId": "cp-res_id_VDU1_CP2_1"
                },
                {
                    "id": "VDU1_CP3-res_id_VDU1_1",
                    "cpdId": "VDU1_CP3",
                    "vnfLinkPortId": "res_id_VDU1_CP3_1"
                },
                {
                    "id": "VDU1_CP4-res_id_VDU1_1",
                    "cpdId": "VDU1_CP4",
                    "vnfLinkPortId": "res_id_VDU1_CP4_1"
                },
                {
                    "id": "VDU1_CP5-res_id_VDU1_1",
                    "cpdId": "VDU1_CP5",
                    "vnfLinkPortId": "res_id_VDU1_CP5_1"
                }
            ],
            # "metadata": omitted
        },
        {
            "id": "res_id_VDU2",
            "vduId": "VDU2",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU2",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "vnfcCpInfo": [
                {
                    "id": "VDU2_CP1-res_id_VDU2",
                    "cpdId": "VDU2_CP1",
                    "vnfExtCpId": "cp-res_id_VDU2_CP1"
                },
                {
                    "id": "VDU2_CP2-res_id_VDU2",
                    "cpdId": "VDU2_CP2",
                    "vnfExtCpId": "cp-res_id_VDU2_CP2"
                },
                {
                    "id": "VDU2_CP3-res_id_VDU2",
                    "cpdId": "VDU2_CP3",
                    "vnfLinkPortId": "res_id_VDU2_CP3"
                },
                {
                    "id": "VDU2_CP4-res_id_VDU2",
                    "cpdId": "VDU2_CP4",
                    "vnfLinkPortId": "res_id_VDU2_CP4"
                },
                {
                    "id": "VDU2_CP5-res_id_VDU2",
                    "cpdId": "VDU2_CP5",
                    "vnfLinkPortId": "res_id_VDU2_CP5"
                }
            ],
            # "metadata": omitted
        }
    ],
    "vnfVirtualLinkResourceInfo": [
        {
            "id": "res_id_internalVL3",
            "vnfVirtualLinkDescId": "internalVL3",
            "networkResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_internalVL3",
                "vimLevelResourceType": "OS::Neutron::Net"
            },
            "vnfLinkPorts": [
                {
                    "id": "res_id_VDU2_CP5",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU2_CP5",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU2_CP5-res_id_VDU2",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "res_id_VDU1_CP5_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP5_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU1_CP5-res_id_VDU1_1",
                    "cpInstanceType": "VNFC_CP"
                }
            ]
        },
        {
            "id": "res_id_internalVL2",
            "vnfVirtualLinkDescId": "internalVL2",
            "networkResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_internalVL2",
                "vimLevelResourceType": "OS::Neutron::Net"
            },
            "vnfLinkPorts": [
                {
                    "id": "res_id_VDU2_CP4",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU2_CP4",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU2_CP4-res_id_VDU2",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "res_id_VDU1_CP4_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP4_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU1_CP4-res_id_VDU1_1",
                    "cpInstanceType": "VNFC_CP"
                }
            ]
        }
    ],
    "virtualStorageResourceInfo": [
        {
            "id": "res_id_VirtualStorage_1",
            "virtualStorageDescId": "VirtualStorage",
            "storageResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VirtualStorage_1",
                "vimLevelResourceType": "OS::Cinder::Volume"
            }
        }
    ],
    # "vnfcInfo": omitted
}

# example_2 is added a VDU1 to example_1.
_inst_info_example_2 = {
    # "flavourId", "vnfState", "scaleStatus", "maxScaleLevels" are omitted
    # "extCpInfo": omitted
    "extVirtualLinkInfo": [
        {
            "id": "bbf0932a-6142-4ea8-93cd-8059dba594a1",
            "resourceHandle": {
                "resourceId": "3529d333-dbcc-4d93-9b64-210647712569"
            },
            "extLinkPorts": [
                {
                    "id": "res_id_VDU2_CP1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU2_CP1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU2_CP1"
                },
                {
                    "id": "res_id_VDU1_CP1_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP1_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU1_CP1_1"
                },
                {
                    "id": "res_id_VDU1_CP1_2",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP1_2",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU1_CP1_2"
                }
            ],
            # "currentVnfExtCpData": omitted
        },
        {
            "id": "790949df-c7b3-4926-a559-3895412f1dfe",
            "resourceHandle": {
                "resourceId": "367e5b3b-34dc-47f2-85b8-c39e3272893a"
            },
            "extLinkPorts": [
                {
                    "id": "res_id_VDU2_CP2",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU2_CP2",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU2_CP2"
                },
                {
                    "id": "res_id_VDU1_CP2_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP2_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU1_CP2_1"
                },
                {
                    "id": "res_id_VDU1_CP2_2",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP2_2",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU1_CP2_2"
                }
            ],
            # "currentVnfExtCpData": omitted
        }
    ],
    "extManagedVirtualLinkInfo": [
        {
            "id": "res_id_internalVL1",
            "vnfVirtualLinkDescId": "internalVL1",
            "networkResource": {
                "resourceId": "res_id_internalVL1"
            },
            "vnfLinkPorts": [
                {
                    "id": "res_id_VDU2_CP3",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU2_CP3",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU2_CP3-res_id_VDU2",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "res_id_VDU1_CP3_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP3_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU1_CP3-res_id_VDU1_1",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "res_id_VDU1_CP3_2",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP3_2",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU1_CP3-res_id_VDU1_2",
                    "cpInstanceType": "VNFC_CP"
                }
            ]
        }
    ],
    "vnfcResourceInfo": [
        {
            "id": "res_id_VDU1_2",
            "vduId": "VDU1",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU1_2",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "storageResourceIds": [
                "res_id_VirtualStorage_2"
            ],
            "vnfcCpInfo": [
                {
                    "id": "VDU1_CP1-res_id_VDU1_2",
                    "cpdId": "VDU1_CP1",
                    "vnfExtCpId": "cp-res_id_VDU1_CP1_2"
                },
                {
                    "id": "VDU1_CP2-res_id_VDU1_2",
                    "cpdId": "VDU1_CP2",
                    "vnfExtCpId": "cp-res_id_VDU1_CP2_2"
                },
                {
                    "id": "VDU1_CP3-res_id_VDU1_2",
                    "cpdId": "VDU1_CP3",
                    "vnfLinkPortId": "res_id_VDU1_CP3_2"
                },
                {
                    "id": "VDU1_CP4-res_id_VDU1_2",
                    "cpdId": "VDU1_CP4",
                    "vnfLinkPortId": "res_id_VDU1_CP4_2"
                },
                {
                    "id": "VDU1_CP5-res_id_VDU1_2",
                    "cpdId": "VDU1_CP5",
                    "vnfLinkPortId": "res_id_VDU1_CP5_2"
                }
            ],
            # "metadata": omitted
        },
        {
            "id": "res_id_VDU1_1",
            "vduId": "VDU1",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU1_1",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "storageResourceIds": [
                "res_id_VirtualStorage_1"
            ],
            "vnfcCpInfo": [
                {
                    "id": "VDU1_CP1-res_id_VDU1_1",
                    "cpdId": "VDU1_CP1",
                    "vnfExtCpId": "cp-res_id_VDU1_CP1_1"
                },
                {
                    "id": "VDU1_CP2-res_id_VDU1_1",
                    "cpdId": "VDU1_CP2",
                    "vnfExtCpId": "cp-res_id_VDU1_CP2_1"
                },
                {
                    "id": "VDU1_CP3-res_id_VDU1_1",
                    "cpdId": "VDU1_CP3",
                    "vnfLinkPortId": "res_id_VDU1_CP3_1"
                },
                {
                    "id": "VDU1_CP4-res_id_VDU1_1",
                    "cpdId": "VDU1_CP4",
                    "vnfLinkPortId": "res_id_VDU1_CP4_1"
                },
                {
                    "id": "VDU1_CP5-res_id_VDU1_1",
                    "cpdId": "VDU1_CP5",
                    "vnfLinkPortId": "res_id_VDU1_CP5_1"
                }
            ],
            # "metadata": omitted
        },
        {
            "id": "res_id_VDU2",
            "vduId": "VDU2",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU2",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "vnfcCpInfo": [
                {
                    "id": "VDU2_CP1-res_id_VDU2",
                    "cpdId": "VDU2_CP1",
                    "vnfExtCpId": "cp-res_id_VDU2_CP1"
                },
                {
                    "id": "VDU2_CP2-res_id_VDU2",
                    "cpdId": "VDU2_CP2",
                    "vnfExtCpId": "cp-res_id_VDU2_CP2"
                },
                {
                    "id": "VDU2_CP3-res_id_VDU2",
                    "cpdId": "VDU2_CP3",
                    "vnfLinkPortId": "res_id_VDU2_CP3"
                },
                {
                    "id": "VDU2_CP4-res_id_VDU2",
                    "cpdId": "VDU2_CP4",
                    "vnfLinkPortId": "res_id_VDU2_CP4"
                },
                {
                    "id": "VDU2_CP5-res_id_VDU2",
                    "cpdId": "VDU2_CP5",
                    "vnfLinkPortId": "res_id_VDU2_CP5"
                }
            ],
            # "metadata": omitted
        }
    ],
    "vnfVirtualLinkResourceInfo": [
        {
            "id": "res_id_internalVL3",
            "vnfVirtualLinkDescId": "internalVL3",
            "networkResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_internalVL3",
                "vimLevelResourceType": "OS::Neutron::Net"
            },
            "vnfLinkPorts": [
                {
                    "id": "res_id_VDU2_CP5",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU2_CP5",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU2_CP5-res_id_VDU2",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "res_id_VDU1_CP5_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP5_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU1_CP5-res_id_VDU1_1",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "res_id_VDU1_CP5_2",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP5_2",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU1_CP5-res_id_VDU1_2",
                    "cpInstanceType": "VNFC_CP"
                }
            ]
        },
        {
            "id": "res_id_internalVL2",
            "vnfVirtualLinkDescId": "internalVL2",
            "networkResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_internalVL2",
                "vimLevelResourceType": "OS::Neutron::Net"
            },
            "vnfLinkPorts": [
                {
                    "id": "res_id_VDU2_CP4",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU2_CP4",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU2_CP4-res_id_VDU2",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "res_id_VDU1_CP4_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP4_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU1_CP4-res_id_VDU1_1",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "res_id_VDU1_CP4_2",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP4_2",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU1_CP4-res_id_VDU1_2",
                    "cpInstanceType": "VNFC_CP"
                }
            ]
        }
    ],
    "virtualStorageResourceInfo": [
        {
            "id": "res_id_VirtualStorage_1",
            "virtualStorageDescId": "VirtualStorage",
            "storageResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VirtualStorage_1",
                "vimLevelResourceType": "OS::Cinder::Volume"
            }
        },
        {
            "id": "res_id_VirtualStorage_2",
            "virtualStorageDescId": "VirtualStorage",
            "storageResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VirtualStorage_2",
                "vimLevelResourceType": "OS::Cinder::Volume"
            }
        }
    ],
    # "vnfcInfo": omitted
}

# example_3 is changed external virtual link ports from example_1.
_inst_info_example_3 = {
    # "flavourId", "vnfState", "scaleStatus", "maxScaleLevels" are omitted
    # "extCpInfo": omitted
    "extVirtualLinkInfo": [
        {
            "id": "bbf0932a-6142-4ea8-93cd-8059dba594a1",
            "resourceHandle": {
                "resourceId": "3529d333-dbcc-4d93-9b64-210647712569"
            },
            "extLinkPorts": [
                {
                    "id": "res_id_VDU1_CP1_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP1_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU1_CP1_1"
                }
            ],
            # "currentVnfExtCpData": omitted
        },
        {
            "id": "2ff742fc-da6d-423a-8fba-6aa6af8da6f2",
            "resourceHandle": {
                "resourceId": "9113aff7-9ba2-43f4-8b1e-fff80ae001c5"
            },
            "extLinkPorts": [
                {
                    "id": "res_id_VDU2_CP1_modified",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU2_CP1_modified",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU2_CP1"
                },
            ],
            # "currentVnfExtCpData": omitted
        },
        {
            "id": "790949df-c7b3-4926-a559-3895412f1dfe",
            "resourceHandle": {
                "resourceId": "367e5b3b-34dc-47f2-85b8-c39e3272893a"
            },
            "extLinkPorts": [
                {
                    "id": "res_id_VDU2_CP2",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU2_CP2",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU2_CP2"
                },
                {
                    "id": "res_id_VDU1_CP2_1",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP2_1",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "cp-res_id_VDU1_CP2_1"
                }
            ],
            # "currentVnfExtCpData": omitted
        },
        {  # same as _inst_info_example_1
            "id": "cba2e572-2cc5-4d11-af60-728615883206",
            "resourceHandle": {
                "resourceId": "4529d333-dbcc-4d93-9b64-210647712569"
            },
            "extLinkPorts": [],
            "currentVnfExtCpData": [
                {
                    "cpdId": "VDU1_CP1",
                    "cpConfig": {
                        "VDU1_CP1_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "numDynamicAddresses": 1
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                },
                {
                    "cpdId": "VDU2_CP1",
                    "cpConfig": {
                        "VDU2_CP1_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "fixedAddresses": [
                                                    "10.10.0.102"
                                                ]
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                }
            ]
        },
        {  # ip address of VDU2_CP2 is changed
            "id": "fc131806-8a4f-43a3-82c0-a38e87fe87bd",
            "resourceHandle": {
                "resourceId": "5529d333-dbcc-4d93-9b64-210647712569"
            },
            "extLinkPorts": [],
            "currentVnfExtCpData": [
                {
                    "cpdId": "VDU1_CP2",
                    "cpConfig": {
                        "VDU1_CP2_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "numDynamicAddresses": 1,
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                },
                {
                    "cpdId": "VDU2_CP2",
                    "cpConfig": {
                        "VDU2_CP2_1": {
                            "cpProtocolData": [
                                {
                                    "layerProtocol": "IP_OVER_ETHERNET",
                                    "ipOverEthernet": {
                                        "ipAddresses": [
                                            {
                                                "type": "IPV4",
                                                "fixedAddresses": [
                                                    "10.10.1.103"
                                                ],
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    }
                }
            ]
        }
    ],
    # NOTE: vnfcCpInfo of VnfcResourceInfo is changed exactly, but it is
    # not looked at update_lcmocc.
    "vnfcResourceInfo": _inst_info_example_1["vnfcResourceInfo"],
    # other members are same as example_1
    "extManagedVirtualLinkInfo":
        _inst_info_example_1["extManagedVirtualLinkInfo"],
    "vnfVirtualLinkResourceInfo":
        _inst_info_example_1["vnfVirtualLinkResourceInfo"],
    "virtualStorageResourceInfo":
        _inst_info_example_1["virtualStorageResourceInfo"],
    # "vnfcInfo": omitted
}

# example_4 is for update_lcmocc test in case of heal. based on example_2.
# * VDU1_1: update server only. VDU1_2: update both server and volume.
# * ports of extMangagedVirtualLink are re-created.
# NOTE: combination of the above sentences can not be happened really.
# it is the data for unit test.
_inst_info_example_4 = {
    # "flavourId", "vnfState", "scaleStatus", "maxScaleLevels" are omitted
    # "extCpInfo": omitted
    "extVirtualLinkInfo": _inst_info_example_2["extVirtualLinkInfo"],
    # network resource is not changed but ports are re-receated.
    # this is for check of SOL003 all=True case.
    "extManagedVirtualLinkInfo": [
        {
            "id": "res_id_internalVL1",
            "vnfVirtualLinkDescId": "internalVL1",
            "networkResource": {
                "resourceId": "res_id_internalVL1"
            },
            "vnfLinkPorts": [
                {
                    "id": "res_id_VDU2_CP3_changed",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU2_CP3_changed",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU2_CP3-res_id_VDU2",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "res_id_VDU1_CP3_1_changed",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP3_1_changed",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU1_CP3-res_id_VDU1_1",
                    "cpInstanceType": "VNFC_CP"
                },
                {
                    "id": "res_id_VDU1_CP3_2_changed",
                    "resourceHandle": {
                        "vimConnectionId": "vim_connection_id",
                        "resourceId": "res_id_VDU1_CP3_2_changed",
                        "vimLevelResourceType": "OS::Neutron::Port"
                    },
                    "cpInstanceId": "VDU1_CP3-res_id_VDU1_2",
                    "cpInstanceType": "VNFC_CP"
                }
            ]
        }
    ],
    "vnfcResourceInfo": [
        {
            "id": "res_id_VDU1_2_new",
            "vduId": "VDU1",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU1_2_new",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "storageResourceIds": [
                "res_id_VirtualStorage_2_new"
            ],
            "vnfcCpInfo": [
                {
                    "id": "VDU1_CP1-res_id_VDU1_2_new",
                    "cpdId": "VDU1_CP1",
                    "vnfExtCpId": "cp-res_id_VDU1_CP1_2_new"
                },
                {
                    "id": "VDU1_CP2-res_id_VDU1_2_new",
                    "cpdId": "VDU1_CP2",
                    "vnfExtCpId": "cp-res_id_VDU1_CP2_2_new"
                },
                {
                    "id": "VDU1_CP3-res_id_VDU1_2_new",
                    "cpdId": "VDU1_CP3",
                    "vnfLinkPortId": "res_id_VDU1_CP3_2_new"
                },
                {
                    "id": "VDU1_CP4-res_id_VDU1_2_new",
                    "cpdId": "VDU1_CP4",
                    "vnfLinkPortId": "res_id_VDU1_CP4_2_new"
                },
                {
                    "id": "VDU1_CP5-res_id_VDU1_2_new",
                    "cpdId": "VDU1_CP5",
                    "vnfLinkPortId": "res_id_VDU1_CP5_2_new"
                }
            ],
            # "metadata": omitted
        },
        {
            "id": "res_id_VDU1_1_new",
            "vduId": "VDU1",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU1_1_new",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "storageResourceIds": [
                "res_id_VirtualStorage_1"
            ],
            "vnfcCpInfo": [
                {
                    "id": "VDU1_CP1-res_id_VDU1_1_new",
                    "cpdId": "VDU1_CP1",
                    "vnfExtCpId": "cp-res_id_VDU1_CP1_1_new"
                },
                {
                    "id": "VDU1_CP2-res_id_VDU1_1_new",
                    "cpdId": "VDU1_CP2",
                    "vnfExtCpId": "cp-res_id_VDU1_CP2_1_new"
                },
                {
                    "id": "VDU1_CP3-res_id_VDU1_1_new",
                    "cpdId": "VDU1_CP3",
                    "vnfLinkPortId": "res_id_VDU1_CP3_1_new"
                },
                {
                    "id": "VDU1_CP4-res_id_VDU1_1_new",
                    "cpdId": "VDU1_CP4",
                    "vnfLinkPortId": "res_id_VDU1_CP4_1_new"
                },
                {
                    "id": "VDU1_CP5-res_id_VDU1_1_new",
                    "cpdId": "VDU1_CP5",
                    "vnfLinkPortId": "res_id_VDU1_CP5_1_new"
                }
            ],
            # "metadata": omitted
        },
        {
            "id": "res_id_VDU2",
            "vduId": "VDU2",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU2",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "vnfcCpInfo": [
                {
                    "id": "VDU2_CP1-res_id_VDU2",
                    "cpdId": "VDU2_CP1",
                    "vnfExtCpId": "cp-res_id_VDU2_CP1"
                },
                {
                    "id": "VDU2_CP2-res_id_VDU2",
                    "cpdId": "VDU2_CP2",
                    "vnfExtCpId": "cp-res_id_VDU2_CP2"
                },
                {
                    "id": "VDU2_CP3-res_id_VDU2",
                    "cpdId": "VDU2_CP3",
                    "vnfLinkPortId": "res_id_VDU2_CP3"
                },
                {
                    "id": "VDU2_CP4-res_id_VDU2",
                    "cpdId": "VDU2_CP4",
                    "vnfLinkPortId": "res_id_VDU2_CP4"
                },
                {
                    "id": "VDU2_CP5-res_id_VDU2",
                    "cpdId": "VDU2_CP5",
                    "vnfLinkPortId": "res_id_VDU2_CP5"
                }
            ],
            # "metadata": omitted
        }
    ],
    "vnfVirtualLinkResourceInfo":
        _inst_info_example_2["vnfVirtualLinkResourceInfo"],
    "virtualStorageResourceInfo": [
        {
            "id": "res_id_VirtualStorage_1",
            "virtualStorageDescId": "VirtualStorage",
            "storageResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VirtualStorage_1",
                "vimLevelResourceType": "OS::Cinder::Volume"
            }
        },
        {
            "id": "res_id_VirtualStorage_2_new",
            "virtualStorageDescId": "VirtualStorage",
            "storageResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VirtualStorage_2_new",
                "vimLevelResourceType": "OS::Cinder::Volume"
            }
        }
    ],
    # "vnfcInfo": omitted
}

# expected results
_expected_resource_changes_instantiate = {
    "affectedVnfcs": [
        {
            "id": "res_id_VDU1_1",
            "vduId": "VDU1",
            "changeType": "ADDED",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU1_1",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "affectedVnfcCpIds": [
                "VDU1_CP1-res_id_VDU1_1",
                "VDU1_CP2-res_id_VDU1_1",
                "VDU1_CP3-res_id_VDU1_1",
                "VDU1_CP4-res_id_VDU1_1",
                "VDU1_CP5-res_id_VDU1_1"
            ],
            "addedStorageResourceIds": [
                "res_id_VirtualStorage_1"
            ]
        },
        {
            "id": "res_id_VDU2",
            "vduId": "VDU2",
            "changeType": "ADDED",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU2",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "affectedVnfcCpIds": [
                "VDU2_CP1-res_id_VDU2",
                "VDU2_CP2-res_id_VDU2",
                "VDU2_CP3-res_id_VDU2",
                "VDU2_CP4-res_id_VDU2",
                "VDU2_CP5-res_id_VDU2"
            ]
        }
    ],
    "affectedVirtualLinks": [
        {
            "id": "res_id_internalVL1",
            "vnfVirtualLinkDescId": "internalVL1",
            "changeType": "LINK_PORT_ADDED",
            "networkResource": {
                "resourceId": "res_id_internalVL1"
            },
            "vnfLinkPortIds": [
                "res_id_VDU1_CP3_1",
                "res_id_VDU2_CP3"
            ]
        },
        {
            "id": "res_id_internalVL2",
            "vnfVirtualLinkDescId": "internalVL2",
            "changeType": "ADDED",
            "networkResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_internalVL2",
                "vimLevelResourceType": "OS::Neutron::Net"
            },
            "vnfLinkPortIds": [
                "res_id_VDU1_CP4_1",
                "res_id_VDU2_CP4"
            ]
        },
        {
            "id": "res_id_internalVL3",
            "vnfVirtualLinkDescId": "internalVL3",
            "changeType": "ADDED",
            "networkResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_internalVL3",
                "vimLevelResourceType": "OS::Neutron::Net"
            },
            "vnfLinkPortIds": [
                "res_id_VDU1_CP5_1",
                "res_id_VDU2_CP5"
            ]
        }
    ],
    "affectedExtLinkPorts": [
        {
            "id": "res_id_VDU1_CP1_1",
            "changeType": "ADDED",
            "extCpInstanceId": "cp-res_id_VDU1_CP1_1",
            "resourceHandle": {
                "resourceId": "res_id_VDU1_CP1_1",
                "vimConnectionId": "vim_connection_id",
                "vimLevelResourceType": "OS::Neutron::Port"
            }
        },
        {
            "id": "res_id_VDU1_CP2_1",
            "changeType": "ADDED",
            "extCpInstanceId": "cp-res_id_VDU1_CP2_1",
            "resourceHandle": {
                "resourceId": "res_id_VDU1_CP2_1",
                "vimConnectionId": "vim_connection_id",
                "vimLevelResourceType": "OS::Neutron::Port"
            }
        },
        {
            "id": "res_id_VDU2_CP1",
            "changeType": "ADDED",
            "extCpInstanceId": "cp-res_id_VDU2_CP1",
            "resourceHandle": {
                "resourceId": "res_id_VDU2_CP1",
                "vimConnectionId": "vim_connection_id",
                "vimLevelResourceType": "OS::Neutron::Port"
            }
        },
        {
            "id": "res_id_VDU2_CP2",
            "changeType": "ADDED",
            "extCpInstanceId": "cp-res_id_VDU2_CP2",
            "resourceHandle": {
                "resourceId": "res_id_VDU2_CP2",
                "vimConnectionId": "vim_connection_id",
                "vimLevelResourceType": "OS::Neutron::Port"
            }
        },
    ],
    "affectedVirtualStorages": [
        {
            "id": "res_id_VirtualStorage_1",
            "virtualStorageDescId": "VirtualStorage",
            "changeType": "ADDED",
            "storageResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VirtualStorage_1",
                "vimLevelResourceType": "OS::Cinder::Volume"
            }
        }
    ]
}

_expected_resource_changes_scale_out = {
    "affectedVnfcs": [
        {
            "id": "res_id_VDU1_2",
            "vduId": "VDU1",
            "changeType": "ADDED",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU1_2",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "affectedVnfcCpIds": [
                "VDU1_CP1-res_id_VDU1_2",
                "VDU1_CP2-res_id_VDU1_2",
                "VDU1_CP3-res_id_VDU1_2",
                "VDU1_CP4-res_id_VDU1_2",
                "VDU1_CP5-res_id_VDU1_2"
            ],
            "addedStorageResourceIds": [
                "res_id_VirtualStorage_2"
            ]
        }
    ],
    "affectedVirtualLinks": [
        {
            "id": "res_id_internalVL1",
            "vnfVirtualLinkDescId": "internalVL1",
            "changeType": "LINK_PORT_ADDED",
            "networkResource": {
                "resourceId": "res_id_internalVL1"
            },
            "vnfLinkPortIds": [
                "res_id_VDU1_CP3_2"
            ]
        },
        {
            "id": "res_id_internalVL2",
            "vnfVirtualLinkDescId": "internalVL2",
            "changeType": "LINK_PORT_ADDED",
            "networkResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_internalVL2",
                "vimLevelResourceType": "OS::Neutron::Net"
            },
            "vnfLinkPortIds": [
                "res_id_VDU1_CP4_2"
            ]
        },
        {
            "id": "res_id_internalVL3",
            "vnfVirtualLinkDescId": "internalVL3",
            "changeType": "LINK_PORT_ADDED",
            "networkResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_internalVL3",
                "vimLevelResourceType": "OS::Neutron::Net"
            },
            "vnfLinkPortIds": [
                "res_id_VDU1_CP5_2"
            ]
        }
    ],
    "affectedExtLinkPorts": [
        {
            "id": "res_id_VDU1_CP1_2",
            "changeType": "ADDED",
            "extCpInstanceId": "cp-res_id_VDU1_CP1_2",
            "resourceHandle": {
                "resourceId": "res_id_VDU1_CP1_2",
                "vimConnectionId": "vim_connection_id",
                "vimLevelResourceType": "OS::Neutron::Port"
            }
        },
        {
            "id": "res_id_VDU1_CP2_2",
            "changeType": "ADDED",
            "extCpInstanceId": "cp-res_id_VDU1_CP2_2",
            "resourceHandle": {
                "resourceId": "res_id_VDU1_CP2_2",
                "vimConnectionId": "vim_connection_id",
                "vimLevelResourceType": "OS::Neutron::Port"
            }
        }
    ],
    "affectedVirtualStorages": [
        {
            "id": "res_id_VirtualStorage_2",
            "virtualStorageDescId": "VirtualStorage",
            "changeType": "ADDED",
            "storageResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VirtualStorage_2",
                "vimLevelResourceType": "OS::Cinder::Volume"
            }
        }
    ]
}

_expected_resource_changes_scale_in = {
    "affectedVnfcs": [
        {
            "id": "res_id_VDU1_2",
            "vduId": "VDU1",
            "changeType": "REMOVED",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU1_2",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "affectedVnfcCpIds": [
                "VDU1_CP1-res_id_VDU1_2",
                "VDU1_CP2-res_id_VDU1_2",
                "VDU1_CP3-res_id_VDU1_2",
                "VDU1_CP4-res_id_VDU1_2",
                "VDU1_CP5-res_id_VDU1_2"
            ],
            "removedStorageResourceIds": [
                "res_id_VirtualStorage_2"
            ]
        }
    ],
    "affectedVirtualLinks": [
        {
            "id": "res_id_internalVL1",
            "vnfVirtualLinkDescId": "internalVL1",
            "changeType": "LINK_PORT_REMOVED",
            "networkResource": {
                "resourceId": "res_id_internalVL1"
            },
            "vnfLinkPortIds": [
                "res_id_VDU1_CP3_2"
            ]
        },
        {
            "id": "res_id_internalVL2",
            "vnfVirtualLinkDescId": "internalVL2",
            "changeType": "LINK_PORT_REMOVED",
            "networkResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_internalVL2",
                "vimLevelResourceType": "OS::Neutron::Net"
            },
            "vnfLinkPortIds": [
                "res_id_VDU1_CP4_2"
            ]
        },
        {
            "id": "res_id_internalVL3",
            "vnfVirtualLinkDescId": "internalVL3",
            "changeType": "LINK_PORT_REMOVED",
            "networkResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_internalVL3",
                "vimLevelResourceType": "OS::Neutron::Net"
            },
            "vnfLinkPortIds": [
                "res_id_VDU1_CP5_2"
            ]
        }
    ],
    "affectedExtLinkPorts": [
        {
            "id": "res_id_VDU1_CP1_2",
            "changeType": "REMOVED",
            "extCpInstanceId": "cp-res_id_VDU1_CP1_2",
            "resourceHandle": {
                "resourceId": "res_id_VDU1_CP1_2",
                "vimConnectionId": "vim_connection_id",
                "vimLevelResourceType": "OS::Neutron::Port"
            }
        },
        {
            "id": "res_id_VDU1_CP2_2",
            "changeType": "REMOVED",
            "extCpInstanceId": "cp-res_id_VDU1_CP2_2",
            "resourceHandle": {
                "resourceId": "res_id_VDU1_CP2_2",
                "vimConnectionId": "vim_connection_id",
                "vimLevelResourceType": "OS::Neutron::Port"
            }
        }
    ],
    "affectedVirtualStorages": [
        {
            "id": "res_id_VirtualStorage_2",
            "virtualStorageDescId": "VirtualStorage",
            "changeType": "REMOVED",
            "storageResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VirtualStorage_2",
                "vimLevelResourceType": "OS::Cinder::Volume"
            }
        }
    ]
}

_expected_resource_changes_terminate = {
    "affectedVnfcs": [
        {
            "id": "res_id_VDU1_1",
            "vduId": "VDU1",
            "changeType": "REMOVED",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU1_1",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "affectedVnfcCpIds": [
                "VDU1_CP1-res_id_VDU1_1",
                "VDU1_CP2-res_id_VDU1_1",
                "VDU1_CP3-res_id_VDU1_1",
                "VDU1_CP4-res_id_VDU1_1",
                "VDU1_CP5-res_id_VDU1_1"
            ],
            "removedStorageResourceIds": [
                "res_id_VirtualStorage_1"
            ]
        },
        {
            "id": "res_id_VDU2",
            "vduId": "VDU2",
            "changeType": "REMOVED",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU2",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "affectedVnfcCpIds": [
                "VDU2_CP1-res_id_VDU2",
                "VDU2_CP2-res_id_VDU2",
                "VDU2_CP3-res_id_VDU2",
                "VDU2_CP4-res_id_VDU2",
                "VDU2_CP5-res_id_VDU2"
            ]
        }
    ],
    "affectedVirtualLinks": [
        {
            "id": "res_id_internalVL1",
            "vnfVirtualLinkDescId": "internalVL1",
            "changeType": "LINK_PORT_REMOVED",
            "networkResource": {
                "resourceId": "res_id_internalVL1"
            },
            "vnfLinkPortIds": [
                "res_id_VDU1_CP3_1",
                "res_id_VDU2_CP3"
            ]
        },
        {
            "id": "res_id_internalVL2",
            "vnfVirtualLinkDescId": "internalVL2",
            "changeType": "REMOVED",
            "networkResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_internalVL2",
                "vimLevelResourceType": "OS::Neutron::Net"
            },
            "vnfLinkPortIds": [
                "res_id_VDU1_CP4_1",
                "res_id_VDU2_CP4"
            ]
        },
        {
            "id": "res_id_internalVL3",
            "vnfVirtualLinkDescId": "internalVL3",
            "changeType": "REMOVED",
            "networkResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_internalVL3",
                "vimLevelResourceType": "OS::Neutron::Net"
            },
            "vnfLinkPortIds": [
                "res_id_VDU1_CP5_1",
                "res_id_VDU2_CP5"
            ]
        }
    ],
    "affectedExtLinkPorts": [
        {
            "id": "res_id_VDU1_CP1_1",
            "changeType": "REMOVED",
            "extCpInstanceId": "cp-res_id_VDU1_CP1_1",
            "resourceHandle": {
                "resourceId": "res_id_VDU1_CP1_1",
                "vimConnectionId": "vim_connection_id",
                "vimLevelResourceType": "OS::Neutron::Port"
            }
        },
        {
            "id": "res_id_VDU1_CP2_1",
            "changeType": "REMOVED",
            "extCpInstanceId": "cp-res_id_VDU1_CP2_1",
            "resourceHandle": {
                "resourceId": "res_id_VDU1_CP2_1",
                "vimConnectionId": "vim_connection_id",
                "vimLevelResourceType": "OS::Neutron::Port"
            }
        },
        {
            "id": "res_id_VDU2_CP1",
            "changeType": "REMOVED",
            "extCpInstanceId": "cp-res_id_VDU2_CP1",
            "resourceHandle": {
                "resourceId": "res_id_VDU2_CP1",
                "vimConnectionId": "vim_connection_id",
                "vimLevelResourceType": "OS::Neutron::Port"
            }
        },
        {
            "id": "res_id_VDU2_CP2",
            "changeType": "REMOVED",
            "extCpInstanceId": "cp-res_id_VDU2_CP2",
            "resourceHandle": {
                "resourceId": "res_id_VDU2_CP2",
                "vimConnectionId": "vim_connection_id",
                "vimLevelResourceType": "OS::Neutron::Port"
            }
        }
    ],
    "affectedVirtualStorages": [
        {
            "id": "res_id_VirtualStorage_1",
            "virtualStorageDescId": "VirtualStorage",
            "changeType": "REMOVED",
            "storageResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VirtualStorage_1",
                "vimLevelResourceType": "OS::Cinder::Volume"
            }
        }
    ]
}

# update_lcmocc  modifies an "Individual VNF instance" example
_modify_inst_saved_example = {
    "id": "1098e2dc-d954-484e-b417-e594ff03c55b",
    "vnfInstanceName": "instance_name",
    "vnfInstanceDescription": "description",
    "vnfdId": "a93b7f96-f0e1-49f2-b3f0-75f5b4a94d0f",
    "vnfProvider": "provider",
    "vnfProductName": "product name",
    "vnfSoftwareVersion": "software version",
    "vnfdVersion": "vnfd version",
    "instantiationState": "INSTANTIATED",
    "vnfConfigurableProperties": {
        "vnfproperties": "example"
    },
    "metadata": {
        "metadata": "example"
    },
    "extensions": {
        "extensions": "example"
    },
    "vimConnectionInfo": {
        "vim1": {
            "vimType": "ETSINFV.OPENSTACK_KEYSTONE.V_3",
            "vimId": "ca925611-f020-4b0e-a56d-3fd5c6d5bc3d",
            "interfaceInfo": {"endpoint": "http://localhost/identity/v3"},
            "accessInfo": {
                "username": "nfv_user",
                "region": "RegionOne",
                "password": "devstack",
                "project": "nfv",
                "projectDomain": "Default",
                "userDomain": "Default"
            },
            "extra": {
                "key": "value"
            }
        }
    },
    "instantiatedVnfInfo": {
        "vnfcInfo": [
            {
                "id": "VDU1-vnfc_res_info_id_VDU1",
                "vnfcConfigurableProperties": {"key": "value"}
            },
            {
                "id": "VDU2-vnfc_res_info_id_VDU2",
                "vnfcConfigurableProperties": {"key": "value"}
            }
        ]
    }
}

_modify_inst_example = {
    "id": "1098e2dc-d954-484e-b417-e594ff03c55b",
    "vnfInstanceName": "instance_name_1",
    "vnfdId": "a93b7f96-f0e1-49f2-b3f0-75f5b4a94dff",
    "vnfProvider": "provider",
    "vnfProductName": "product name",
    "vnfSoftwareVersion": "software version_1",
    "vnfdVersion": "vnfd version_1",
    "metadata": {
        "metadata": "example_1"
    },
    "extensions": {
        "extensions": "example"
    },
    "vimConnectionInfo": {
        "vim1": {
            "vimType": "ETSINFV.OPENSTACK_KEYSTONE.V_3",
            "vimId": "ca925611-f020-4b0e-a56d-3fd5c6d5bc3d",
            "interfaceInfo": {"endpoint": "http://localhost/identity/v3"},
            "accessInfo": {
                "username": "nfv_user",
                "region": "RegionOne",
                "password": "devstack",
                "project": "nfv",
                "projectDomain": "Default",
                "userDomain": "Default"
            },
            "extra": {
                "key": "value",
                "key_add": "value"
            }
        }
    },
    "instantiatedVnfInfo": {
        "vnfcInfo": [
            {
                "id": "VDU1-vnfc_res_info_id_VDU1",
                "vnfcConfigurableProperties": {
                    "key": "value_mod",
                    "key_add": "value"
                }
            },
            {
                "id": "VDU2-vnfc_res_info_id_VDU2",
                "vnfcConfigurableProperties": {}
            }
        ]
    }
}

_expected_changedInfo = {
    "vnfInstanceName": "instance_name_1",
    "vnfdId": "a93b7f96-f0e1-49f2-b3f0-75f5b4a94dff",
    "vnfSoftwareVersion": "software version_1",
    "vnfdVersion": "vnfd version_1",
    "metadata": {
        "metadata": "example_1"
    },
    "vimConnectionInfo": {
        "vim1": {
            "vimType": "ETSINFV.OPENSTACK_KEYSTONE.V_3",
            "vimId": "ca925611-f020-4b0e-a56d-3fd5c6d5bc3d",
            "interfaceInfo": {"endpoint": "http://localhost/identity/v3"},
            "accessInfo": {
                "username": "nfv_user",
                "region": "RegionOne",
                "password": "devstack",
                "project": "nfv",
                "projectDomain": "Default",
                "userDomain": "Default"
            },
            "extra": {
                "key": "value",
                "key_add": "value"
            }
        }
    },
    "vnfcInfoModifications": [
        {
            "id": "VDU1-vnfc_res_info_id_VDU1",
            "vnfcConfigurableProperties": {
                "key": "value_mod",
                "key_add": "value"
            }
        },
        {
            "id": "VDU2-vnfc_res_info_id_VDU2",
            "vnfcConfigurableProperties": {}
        }
    ]
}

_expected_resource_changes_change_ext_conn = {
    "affectedExtLinkPorts": [
        {
            "id": "res_id_VDU2_CP1",
            "changeType": "REMOVED",
            "extCpInstanceId": "cp-res_id_VDU2_CP1",
            "resourceHandle": {
                "resourceId": "res_id_VDU2_CP1",
                "vimConnectionId": "vim_connection_id",
                "vimLevelResourceType": "OS::Neutron::Port"
            }
        },
        {
            "id": "res_id_VDU2_CP1_modified",
            "changeType": "ADDED",
            "extCpInstanceId": "cp-res_id_VDU2_CP1",
            "resourceHandle": {
                "resourceId": "res_id_VDU2_CP1_modified",
                "vimConnectionId": "vim_connection_id",
                "vimLevelResourceType": "OS::Neutron::Port"
            }
        }
    ]
}

_expected_changed_ext_connectivity = [
    # sort by id
    _inst_info_example_3['extVirtualLinkInfo'][1],
    _inst_info_example_3['extVirtualLinkInfo'][0],
    _inst_info_example_3['extVirtualLinkInfo'][4]
]

_expected_resource_changes_heal = {
    "affectedVnfcs": [
        {
            "id": "res_id_VDU1_1",
            "vduId": "VDU1",
            "changeType": "REMOVED",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU1_1",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "affectedVnfcCpIds": [
                "VDU1_CP1-res_id_VDU1_1",
                "VDU1_CP2-res_id_VDU1_1",
                "VDU1_CP3-res_id_VDU1_1",
                "VDU1_CP4-res_id_VDU1_1",
                "VDU1_CP5-res_id_VDU1_1"
            ]
        },
        {
            "id": "res_id_VDU1_1_new",
            "vduId": "VDU1",
            "changeType": "ADDED",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU1_1_new",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "affectedVnfcCpIds": [
                "VDU1_CP1-res_id_VDU1_1_new",
                "VDU1_CP2-res_id_VDU1_1_new",
                "VDU1_CP3-res_id_VDU1_1_new",
                "VDU1_CP4-res_id_VDU1_1_new",
                "VDU1_CP5-res_id_VDU1_1_new"
            ]
        },
        {
            "id": "res_id_VDU1_2",
            "vduId": "VDU1",
            "changeType": "REMOVED",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU1_2",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "affectedVnfcCpIds": [
                "VDU1_CP1-res_id_VDU1_2",
                "VDU1_CP2-res_id_VDU1_2",
                "VDU1_CP3-res_id_VDU1_2",
                "VDU1_CP4-res_id_VDU1_2",
                "VDU1_CP5-res_id_VDU1_2"
            ],
            "removedStorageResourceIds": [
                "res_id_VirtualStorage_2"
            ]
        },
        {
            "id": "res_id_VDU1_2_new",
            "vduId": "VDU1",
            "changeType": "ADDED",
            "computeResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VDU1_2_new",
                "vimLevelResourceType": "OS::Nova::Server"
            },
            "affectedVnfcCpIds": [
                "VDU1_CP1-res_id_VDU1_2_new",
                "VDU1_CP2-res_id_VDU1_2_new",
                "VDU1_CP3-res_id_VDU1_2_new",
                "VDU1_CP4-res_id_VDU1_2_new",
                "VDU1_CP5-res_id_VDU1_2_new"
            ],
            "addedStorageResourceIds": [
                "res_id_VirtualStorage_2_new"
            ]
        },
    ],
    "affectedVirtualLinks": [
        # NOTE: id is same but expected LINK_PORT_ADDED is first based on the
        # code.
        {
            "id": "res_id_internalVL1",
            "vnfVirtualLinkDescId": "internalVL1",
            "changeType": "LINK_PORT_ADDED",
            "networkResource": {
                "resourceId": "res_id_internalVL1"
            },
            "vnfLinkPortIds": [
                "res_id_VDU1_CP3_1_changed",
                "res_id_VDU1_CP3_2_changed",
                "res_id_VDU2_CP3_changed"
            ]
        },
        {
            "id": "res_id_internalVL1",
            "vnfVirtualLinkDescId": "internalVL1",
            "changeType": "LINK_PORT_REMOVED",
            "networkResource": {
                "resourceId": "res_id_internalVL1"
            },
            "vnfLinkPortIds": [
                "res_id_VDU1_CP3_1",
                "res_id_VDU1_CP3_2",
                "res_id_VDU2_CP3"
            ]
        }
    ],
    "affectedVirtualStorages": [
        {
            "id": "res_id_VirtualStorage_2",
            "virtualStorageDescId": "VirtualStorage",
            "changeType": "REMOVED",
            "storageResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VirtualStorage_2",
                "vimLevelResourceType": "OS::Cinder::Volume"
            }
        },
        {
            "id": "res_id_VirtualStorage_2_new",
            "virtualStorageDescId": "VirtualStorage",
            "changeType": "ADDED",
            "storageResource": {
                "vimConnectionId": "vim_connection_id",
                "resourceId": "res_id_VirtualStorage_2_new",
                "vimLevelResourceType": "OS::Cinder::Volume"
            }
        }
    ]
}


class TestLcmOpOccUtils(base.BaseTestCase):

    def setUp(self):
        super(TestLcmOpOccUtils, self).setUp()
        objects.register_all()
        self.context = context.get_admin_context()

    def _sort_resource_changes(self, result):
        # sort lists before compare with an expected_result since
        # order of list items is unpredictable.
        # note that an expected_result is already sorted.
        def _get_key(obj):
            return obj['id']

        if 'affectedVnfcs' in result:
            result['affectedVnfcs'].sort(key=_get_key)
            for vnfc in result['affectedVnfcs']:
                if 'affectedVnfcCpIds' in vnfc:
                    vnfc['affectedVnfcCpIds'].sort()
                if 'removedStorageResourceIds' in vnfc:
                    vnfc['removedStorageResourceIds'].sort()

        if 'affectedVirtualLinks' in result:
            result['affectedVirtualLinks'].sort(key=_get_key)
            for vl in result['affectedVirtualLinks']:
                if 'vnfLinkPortIds' in vl:
                    vl['vnfLinkPortIds'].sort()

        if 'affectedExtLinkPorts' in result:
            result['affectedExtLinkPorts'].sort(key=_get_key)

        if 'affectedVirtualStorages' in result:
            result['affectedVirtualStorages'].sort(key=_get_key)

        return result

    def test_update_lcmocc_instantiate(self):
        # prepare
        inst_saved = objects.VnfInstanceV2()
        inst = objects.VnfInstanceV2()
        inst.instantiatedVnfInfo = (
            objects.VnfInstanceV2_InstantiatedVnfInfo.from_dict(
                _inst_info_example_1))
        lcmocc = objects.VnfLcmOpOccV2(
            operation=fields.LcmOperationType.INSTANTIATE)

        # execute update_lcmocc
        lcmocc_utils.update_lcmocc(lcmocc, inst_saved, inst)

        # check resourceChanges
        lcmocc = lcmocc.to_dict()
        self.assertEqual(
            _expected_resource_changes_instantiate,
            self._sort_resource_changes(lcmocc['resourceChanges']))

    def test_update_lcmocc_scale_out(self):
        # prepare
        inst_saved = objects.VnfInstanceV2()
        inst_saved.instantiatedVnfInfo = (
            objects.VnfInstanceV2_InstantiatedVnfInfo.from_dict(
                _inst_info_example_1))
        inst = objects.VnfInstanceV2()
        inst.instantiatedVnfInfo = (
            objects.VnfInstanceV2_InstantiatedVnfInfo.from_dict(
                _inst_info_example_2))
        lcmocc = objects.VnfLcmOpOccV2(
            operation=fields.LcmOperationType.SCALE)

        # execute update_lcmocc
        lcmocc_utils.update_lcmocc(lcmocc, inst_saved, inst)

        # check resourceChanges
        lcmocc = lcmocc.to_dict()
        self.assertEqual(
            _expected_resource_changes_scale_out,
            self._sort_resource_changes(lcmocc['resourceChanges']))

    def test_update_lcmocc_scale_in(self):
        # prepare
        inst_saved = objects.VnfInstanceV2()
        inst_saved.instantiatedVnfInfo = (
            objects.VnfInstanceV2_InstantiatedVnfInfo.from_dict(
                _inst_info_example_2))
        inst = objects.VnfInstanceV2()
        inst.instantiatedVnfInfo = (
            objects.VnfInstanceV2_InstantiatedVnfInfo.from_dict(
                _inst_info_example_1))
        lcmocc = objects.VnfLcmOpOccV2(
            operation=fields.LcmOperationType.SCALE)

        # execute update_lcmocc
        lcmocc_utils.update_lcmocc(lcmocc, inst_saved, inst)

        # check resourceChanges
        lcmocc = lcmocc.to_dict()
        self.assertEqual(_expected_resource_changes_scale_in,
            self._sort_resource_changes(lcmocc['resourceChanges']))

    def test_update_lcmocc_terminate(self):
        # prepare
        inst_saved = objects.VnfInstanceV2()
        inst_saved.instantiatedVnfInfo = (
            objects.VnfInstanceV2_InstantiatedVnfInfo.from_dict(
                _inst_info_example_1))
        inst = objects.VnfInstanceV2()
        inst.instantiatedVnfInfo = objects.VnfInstanceV2_InstantiatedVnfInfo(
            flavourId="SAMPLE_VNFD_ID",
            vnfState='STOPPED')
        lcmocc = objects.VnfLcmOpOccV2(
            operation=fields.LcmOperationType.TERMINATE)

        # execute update_lcmocc
        lcmocc_utils.update_lcmocc(lcmocc, inst_saved, inst)

        # check resourceChanges
        lcmocc = lcmocc.to_dict()
        self.assertEqual(_expected_resource_changes_terminate,
            self._sort_resource_changes(lcmocc['resourceChanges']))

    def test_update_lcmocc_modify(self):
        # prepare
        inst_saved = objects.VnfInstanceV2.from_dict(
            _modify_inst_saved_example)
        inst = objects.VnfInstanceV2.from_dict(_modify_inst_example)
        lcmocc = objects.VnfLcmOpOccV2(
            operation=fields.LcmOperationType.MODIFY_INFO)

        # execute update_lcmocc
        lcmocc_utils.update_lcmocc(lcmocc, inst_saved, inst)
        # check changedInfo
        lcmocc = lcmocc.to_dict()
        self.assertEqual(_expected_changedInfo, lcmocc['changedInfo'])

    def test_update_lcmocc_change_ext_conn(self):
        # prepare
        inst_saved = objects.VnfInstanceV2()
        inst_saved.instantiatedVnfInfo = (
            objects.VnfInstanceV2_InstantiatedVnfInfo.from_dict(
                _inst_info_example_1))
        inst = objects.VnfInstanceV2()
        inst.instantiatedVnfInfo = (
            objects.VnfInstanceV2_InstantiatedVnfInfo.from_dict(
                _inst_info_example_3))
        lcmocc = objects.VnfLcmOpOccV2(
            operation=fields.LcmOperationType.CHANGE_EXT_CONN)

        # execute update_lcmocc
        lcmocc_utils.update_lcmocc(lcmocc, inst_saved, inst)

        # check resourceChanges
        lcmocc = lcmocc.to_dict()
        self.assertEqual(
            _expected_resource_changes_change_ext_conn,
            self._sort_resource_changes(lcmocc['resourceChanges']))
        self.assertEqual(
            _expected_changed_ext_connectivity,
            sorted(lcmocc['changedExtConnectivity'], key=lambda x: x['id']))

    def test_update_lcmocc_heal(self):
        # NOTE: from the view point of update_lcmocc unit test coverage,
        # it is enough to check server and/or volume update.
        # prepare
        inst_saved = objects.VnfInstanceV2()
        inst_saved.instantiatedVnfInfo = (
            objects.VnfInstanceV2_InstantiatedVnfInfo.from_dict(
                _inst_info_example_2))
        inst = objects.VnfInstanceV2()
        inst.instantiatedVnfInfo = (
            objects.VnfInstanceV2_InstantiatedVnfInfo.from_dict(
                _inst_info_example_4))
        lcmocc = objects.VnfLcmOpOccV2(
            operation=fields.LcmOperationType.HEAL)

        # execute update_lcmocc
        lcmocc_utils.update_lcmocc(lcmocc, inst_saved, inst)

        # check resourceChanges
        lcmocc = lcmocc.to_dict()
        self.assertEqual(
            _expected_resource_changes_heal,
            self._sort_resource_changes(lcmocc['resourceChanges']))



# Copyright 2012, VMware, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from unittest import mock

import oslo_i18n

from tacker.agent.linux import utils
from tacker.tests import base


_marker = object()


class AgentUtilsExecuteTest(base.BaseTestCase):
    def setUp(self):
        super(AgentUtilsExecuteTest, self).setUp()
        self.test_file = self.get_temp_file_path('test_execute.tmp')
        open(self.test_file, 'w').close()
        self.process = mock.patch('eventlet.green.subprocess.Popen').start()
        self.process.return_value.returncode = 0
        self.mock_popen = self.process.return_value.communicate

    def test_without_helper(self):
        expected = "%s\n" % self.test_file
        self.mock_popen.return_value = [expected, ""]
        result = utils.execute(["ls", self.test_file])
        self.assertEqual(result, expected)

    def test_with_helper(self):
        expected = "ls %s\n" % self.test_file
        self.mock_popen.return_value = [expected, ""]
        result = utils.execute(["ls", self.test_file], root_helper='echo')
        self.assertEqual(result, expected)

    def test_stderr_true(self):
        expected = "%s\n" % self.test_file
        self.mock_popen.return_value = [expected, ""]
        out = utils.execute(["ls", self.test_file], return_stderr=True)
        self.assertIsInstance(out, tuple)
        self.assertEqual(out, (expected, ""))

    def test_check_exit_code(self):
        self.mock_popen.return_value = ["", ""]
        stdout = utils.execute(["ls", self.test_file[:-1]],
                               check_exit_code=False)
        self.assertEqual("", stdout)

    def test_execute_raises(self):
        self.mock_popen.side_effect = RuntimeError
        self.assertRaises(RuntimeError, utils.execute,
                          ["ls", self.test_file[:-1]])

    def test_process_input(self):
        expected = "%s\n" % self.test_file[:-1]
        self.mock_popen.return_value = [expected, ""]
        result = utils.execute(["cat"], process_input="%s\n" %
                               self.test_file[:-1])
        self.assertEqual(result, expected)

    def test_with_addl_env(self):
        expected = "%s\n" % self.test_file
        self.mock_popen.return_value = [expected, ""]
        result = utils.execute(["ls", self.test_file],
                               addl_env={'foo': 'bar'})
        self.assertEqual(result, expected)

    def test_return_code_log_error_raise_runtime(self):
        self.mock_popen.return_value = ('', '')
        self.process.return_value.returncode = 1
        with mock.patch.object(utils, 'LOG') as log:
            self.assertRaises(RuntimeError, utils.execute,
                              ['ls'])
            self.assertTrue(log.error.called)

    def test_return_code_log_error_no_raise_runtime(self):
        self.mock_popen.return_value = ('', '')
        self.process.return_value.returncode = 1
        with mock.patch.object(utils, 'LOG') as log:
            utils.execute(['ls'], check_exit_code=False)
            self.assertTrue(log.error.called)

    def test_return_code_log_debug(self):
        self.mock_popen.return_value = ('', '')
        with mock.patch.object(utils, 'LOG') as log:
            utils.execute(['ls'])
            self.assertTrue(log.debug.called)

    def test_return_code_log_error_change_locale(self):
        ja_output = 'std_out in Japanese'
        ja_error = 'std_err in Japanese'
        ja_message_out = oslo_i18n._message.Message(ja_output)
        ja_message_err = oslo_i18n._message.Message(ja_error)
        ja_translate_out = oslo_i18n._translate.translate(ja_message_out, 'ja')
        ja_translate_err = oslo_i18n._translate.translate(ja_message_err, 'ja')
        self.mock_popen.return_value = (ja_translate_out, ja_translate_err)
        self.process.return_value.returncode = 1

        with mock.patch.object(utils, 'LOG') as log:
            utils.execute(['ls'], check_exit_code=False)
            self.assertIn(ja_translate_out, str(log.error.call_args_list))
            self.assertIn(ja_translate_err, str(log.error.call_args_list))

    def test_return_code_raise_runtime_log_fail_as_error(self):
        self.mock_popen.return_value = ('', '')
        self.process.return_value.returncode = 1
        with mock.patch.object(utils, 'LOG') as log:
            self.assertRaises(RuntimeError, utils.execute,
                              ['ls'])
            self.assertTrue(log.error.called)

    def test_encode_process_input(self):
        bytes_idata = bytes("%s\n" % self.test_file[:-1], 'utf-8')
        bytes_odata = bytes("%s\n" % self.test_file, 'utf-8')
        self.mock_popen.return_value = [bytes_odata, b'']
        result = utils.execute(['cat'], process_input=bytes_idata)
        self.mock_popen.assert_called_once_with(bytes_idata)
        self.assertEqual(bytes_odata, result)

    def test_return_str_data(self):
        str_data = "%s\n" % self.test_file
        self.mock_popen.return_value = [str_data, '']
        result = utils.execute(['ls', self.test_file], return_stderr=True)
        self.assertEqual((str_data, ''), result)


class AgentUtilsExecuteEncodeTest(base.BaseTestCase):
    def setUp(self):
        super(AgentUtilsExecuteEncodeTest, self).setUp()
        self.test_file = self.get_temp_file_path('test_execute.tmp')
        open(self.test_file, 'w').close()

    def test_decode_return_data(self):
        str_data = bytes("%s\n" % self.test_file, 'utf-8')
        result = utils.execute(['ls', self.test_file], return_stderr=True)
        self.assertEqual((str_data, bytes('', 'utf-8')), result)



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_log import log as logging

from tacker.sol_refactored.common import config
from tacker.sol_refactored.common import http_client
from tacker.sol_refactored.common import lcm_op_occ_utils as lcmocc_utils
from tacker.sol_refactored.common import subscription_utils as subsc_utils
from tacker.sol_refactored.common import vnfd_utils
from tacker.sol_refactored.nfvo import local_nfvo
from tacker.sol_refactored import objects


LOG = logging.getLogger(__name__)

CONF = config.CONF


class NfvoClient(object):

    def __init__(self):
        self.is_local = True
        self.nfvo = local_nfvo.LocalNfvo()

        if CONF.v2_nfvo.use_external_nfvo:
            self.is_local = False
            auth_handle = http_client.OAuth2AuthHandle(
                CONF.v2_nfvo.endpoint,
                CONF.v2_nfvo.token_endpoint,
                CONF.v2_nfvo.client_id,
                CONF.v2_nfvo.client_password)
            self.client = http_client.HttpClient(auth_handle)
            self.grant_api_version = CONF.v2_nfvo.grant_api_version
            self.vnfpkgm_api_version = CONF.v2_nfvo.vnfpkgm_api_version

    def get_vnf_package_info_vnfd(self, context, vnfd_id):
        if self.is_local:
            return self.nfvo.onboarded_show(context, vnfd_id)

        url = "/vnfpkgm/v2/onboarded_vnf_packages/{}".format(vnfd_id)
        resp, body = self.client.do_request(
            url, "GET", expected_status=[200],
            version=self.vnfpkgm_api_version)
        LOG.debug("vnfpkg_info_vnfd: %s" % body)
        return objects.VnfPkgInfoV2.from_dict(body)

    def onboarded_show_vnfd(self, context, vnfd_id):
        if self.is_local:
            # this is not happen. will raise internal server error.
            LOG.error("onboarded_show_vnfd is called.")
            return

        url = "/vnfpkgm/v2/onboarded_vnf_packages/{}/vnfd".format(vnfd_id)
        resp, body = self.client.do_request(
            url, "GET", expected_status=[200],
            version=self.vnfpkgm_api_version)
        return body

    def onboarded_package_content(self, context, vnfd_id):
        if self.is_local:
            # this is not happen. will raise internal server error.
            LOG.error("onboarded_package_content is called.")
            return

        url = "/vnfpkgm/v2/onboarded_vnf_packages/{}/package_content"
        url = url.format(vnfd_id)
        resp, body = self.client.do_request(
            url, "GET", expected_status=[200],
            version=self.vnfpkgm_api_version)
        return body

    def grant(self, context, grant_req):
        LOG.debug("grant request: %s", grant_req.to_dict())

        if self.is_local:
            grant_res = self.nfvo.grant(context, grant_req)
        else:
            url = "/grant/v2/grants"
            resp, body = self.client.do_request(
                url, "POST", expected_status=[201], body=grant_req,
                version=self.grant_api_version)
            grant_res = objects.GrantV1.from_dict(body)

        LOG.debug("grant response: %s", grant_res.to_dict())
        return grant_res

    def get_vnfd(self, context, vnfd_id, all_contents=False):
        if self.is_local:
            return self.nfvo.get_vnfd(context, vnfd_id)

        if all_contents:
            zip_file = self.onboarded_package_content(context, vnfd_id)
        else:
            zip_file = self.onboarded_show_vnfd(context, vnfd_id)

        vnfd = vnfd_utils.Vnfd(vnfd_id)
        vnfd.init_from_zip_file(zip_file)

        return vnfd

    def send_inst_create_notification(self, context, inst, endpoint):
        subscs = subsc_utils.get_inst_create_subscs(context, inst)
        for subsc in subscs:
            notif_data = subsc_utils.make_create_inst_notif_data(
                subsc, inst, endpoint)
            subsc_utils.send_notification(subsc, notif_data)

        if self.is_local:
            self.nfvo.recv_inst_create_notification(context, inst)

    def send_inst_delete_notification(self, context, inst, endpoint):
        subscs = subsc_utils.get_inst_delete_subscs(context, inst)
        for subsc in subscs:
            notif_data = subsc_utils.make_delete_inst_notif_data(
                subsc, inst, endpoint)
            subsc_utils.send_notification(subsc, notif_data)

        if self.is_local:
            self.nfvo.recv_inst_delete_notification(context, inst)

    def send_lcmocc_notification(self, context, lcmocc, inst, endpoint):
        subscs = subsc_utils.get_lcmocc_subscs(context, lcmocc, inst)
        for subsc in subscs:
            notif_data = lcmocc_utils.make_lcmocc_notif_data(
                subsc, lcmocc, endpoint)
            subsc_utils.send_notification(subsc, notif_data)

        if self.is_local:
            self.nfvo.recv_lcmocc_notification(context, lcmocc, inst)
